/*
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:      CCM (obtain.C)
  
  Copyright: Damian J. J. Farnell

  Date:      Mon Feb 21 2000

  Updated:   Joerg Schulenburg 2004-2005

  Purpose:   Determine CCM LSUBm equations.

  ===================================================

  ---------------------------------------------------

  MODULE NAME: OBTAIN.C 

  PURPOSE:     `Obtains' the CCM equations for LSUBm

  ---------------------------------------------------

  Code checked on Thu Aug 3 11:08:54 BST 2000, DJJF.

  ---------------------------------------------------

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "Constants.h"
#include "Externals.h"
#include "FunctionPrototypes.h"

#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

extern void collect_term(int indices[6],double dummy,int id);
   void collect_gse_term(int indices[6],double dummy,int id);
extern int subsetClusters(int (*config)[5], int length); // 1 if ok, 0 if not

int radius=0; // only used within obtain.C, better (faster) locally?
int h_id=-1;  // file id for matrix portion, 0=gse, 1=terms, 2=expects

// defined in collect.C
extern unsigned long
 nterms_in1,  nterms_in2,
 nterms_out1, nterms_out2;

// see at bottom for obtain(void)

//  split configuration conf0 into conf1 and conf2 according to split-array
//  take sites from conf0[1..len0] and append (!) to conf1[len1..] + conf2
// static // extern for obtainGeneralisedExpectation()
int  split_config(int (*conf0)[5], int  len0, int *split0,
                  int (*conf1)[5], int *len1,
                  int (*conf2)[5], int *len2){
  int i, *pd, *ps;
  for (i=1; i<=len0; i++) {
    ps=conf0[i];
    if (split0[i]) { (*len2)++; pd=conf2[*len2]; } // destination config 1
    else           { (*len1)++; pd=conf1[*len1]; } // destination config 0
    pd[1]=ps[1]; // x
    pd[2]=ps[2]; // y
    pd[3]=ps[3]; // z
    pd[4]=ps[4]; // uc
    // pd[5]=ps[5]; Bug until Mar08
  }
  // count up (dual)
  for(i=1;i<=len0;i++) {
    split0[i]^=1; if (split0[i]) break;
  }
  return ((i>len0)?1:0); // counter overflow
}

// Joerg ToDo: bonds or expect is a bad name, better use h_component, Jun08

/* test if site s1 and site s2 are neighbours k -> m */
// static
int test_k_to_m_neighbour(struct hamiltonian_defn hh[],int term, int s1[5], int s2[5]) {
  int are_they_neighbours=0;
  // struct hamiltonian_defn hh[]=&Hamiltonian[bonds][term];
  int index=s1[4];
  int j;
  // neighbourhood of s2 to s1 (can be directional!)
  for (j=1;j<=hh[term].k_to_m_number[index];j++) {
	/* k ---> m                      uc   nn xyzc */
	if (s2[1] == s1[1] + hh[term].k_to_m[index][j][0] /* x */
	 && s2[2] == s1[2] + hh[term].k_to_m[index][j][1] /* y */
	 && s2[3] == s1[3] + hh[term].k_to_m[index][j][2] /* z */
	 && s2[4] ==         hh[term].k_to_m[index][j][3] /* uc */ ) {
             are_they_neighbours=j; break; }
  }
  return are_they_neighbours;
}

// s^z s^z : all s 

double term2(int config_index, int length, int (*string)[5],
             struct hamiltonian_defn hh[], int term)
{
  
  /* 
     TERM 2: 
     
     F(m) . S(+)(m) . S(z)(k)

     Checked 3-7-2000.
     
  */
  
  int i,j,k;
//  double multiplier;
  int (*configuration)[5];
  int neighbour;
  int index;
  int new_index;
  double sum;
  sum=0.0000;
  
  if (!hh[term].evaluate) return 0; /* skip that term */

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];
  
  // get the multiplying coefficient correct first!

  // multiplier=1.0;

  // for(i=1;i<length;i++)      
  // multiplier=multiplier*i; 
  // multiplier=multiplier*length;

  // (length-1)!            -- permutation factor because we sum over m.
  // length                 -- factor from F()

#if DEBUG
  if (Control.output > 10) printf("Term=%d\n",term);
  // why it was inside k-loop? Joerg moved it outside Sep05
  findDescriptorOfCluster(length,configuration,descriptor);
  descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
  if (descriptorIndex != config_index) {
    printf("Term 2 Error: descriptorIndex (%d) != config_index (%d)...\n"
           "Exiting!!\n\n", descriptorIndex, config_index);
    exit(1);
  }
#endif
  
  for(k=1;k<=length;k++) {
    
    index = configuration[k][4]; // for every site ...
    
    for(neighbour=1;
        neighbour<=hh[term].m_to_k_number[index];
	neighbour++) {
      
      /* m ---> k */
      
      new_index = hh[term].m_to_k[index][neighbour][3];
      
      // note that no *CCMcoeffs[config_index] to find in sum
      // because it's the lhs.
      
      // sum+=(-0.5*UnitCell.s2[new_index])*
      //     hh[term].coefficient[index][neighbour];
      
      // -.5*UnitCell.s2[new_index] -- coefficient from \tilde s^z \tilde s^z
      
      if (Control.file) {
	int indices[7]={0};
	double dummy;
	indices[0]=config_index;
        indices[1]=2;
	indices[2]=config_index;
        indices[3]=0;
	indices[4]=0;
        indices[5]=0;
	dummy=(-0.5*UnitCell.s2[new_index])*
	  hh[term].coefficient[index][neighbour];
        collect_term(indices,dummy,h_id);
      }
      
    }
  }
  free(configuration);
  return sum;
}

double term3(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  
  /* 

     TERM 3: 
     
     F(k) . S(+)(k) . S(z)(m)

  */
  
  int i,j,k;
  // double multiplier;
  int (*configuration)[5];
  int neighbour;
  int index;
  int new_index;
  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  // multiplier=1.0;

  // for(i=1;i<length;i++) 
  // multiplier=multiplier*i;
  // multiplier=multiplier*length; 

  // (length-1)!            -- permutation factor because we sum over k.
  // length                 -- factor from F()

#if DEBUG
  int descriptorIndex; // not used, just a check
  // moved outside k-loop by Joerg Sep05
  findDescriptorOfCluster(length,configuration,descriptor);
  descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
  
  if (descriptorIndex != config_index) {
    printf("Term 2 Error: descriptorIndex (%d) != config_index (%d)...\n"
           "Exiting!!\n\n", descriptorIndex, config_index);
    exit(1);
  }
#endif

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];      

    for(neighbour=1;
        neighbour<=hh[term].k_to_m_number[index];
        neighbour++) {
      
      /* k ---> m */
      
      new_index=hh[term].k_to_m[index][neighbour][3];
      
      // note that no *CCMcoeffs[config_index] to find in sum
      // because it's the lhs.
      
      sum+=(-0.5*UnitCell.s2[new_index])*
	hh[term].coefficient[index][neighbour];
      
      // -0.5*UnitCell.s2[new_index] -- coefficient from \tilde s^z \tilde s^z
      
      if (Control.file) {
	int indices[7]={0};
	double dummy;
	indices[0]=config_index;
        indices[1]=3;
	indices[2]=config_index;
        indices[3]=0;
	indices[4]=0;
        indices[5]=0;
	dummy=(-0.5*UnitCell.s2[new_index])*
	  hh[term].coefficient[index][neighbour];
        collect_term(indices,dummy,h_id);
      }
      
    }
  }
  free(configuration);
  return sum;
}

double term4(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 4: 

     H(k,m) . S(+)(k) . S(+)(m).

  */

  int conf[4][6];
  int i, j, k, l;
  int k1;
  int k2;
  //  int nn[5];
  int are_they_neighbours;
  int index;
  //  double multiplier;
  int (*configuration)[5];
  double sum;

  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<3;k++) {
    for(l=1;l<5;l++) {
      conf[k][l]=0;
    }
  } 
  
  // multiplier=1.0;
  
  // for(i=1;i<=(length-2);i++)
  // multiplier=multiplier*i; 
  // multiplier=multiplier*length*(length-1);
  
  // (length-2)!       -- permutation factor because we sum over k1 and k2.
  // length*(length-1) -- factor from F()

  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {  // ToDo: better k1 < k2 and factor 2 ??? Joerg
	
	for(l=1;l<5;l++) { conf[1][l]=configuration[k1][l]; }
	for(l=1;l<5;l++) { conf[2][l]=configuration[k2][l]; }
	
	// loop over nearest neighbours
	
        index = conf[1][4];
        are_they_neighbours=
          test_k_to_m_neighbour(hh, term, conf[1], conf[2]);
	
	// conf[1] and conf[2] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  // a check here...
	  
#if DEBUG
	  findDescriptorOfCluster(length,configuration,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  if (descriptorIndex != config_index) {
      printf("Term 4 Error: descriptorIndex (%d) != config_index (%d)...\n"
             "Exiting!!\n\n", descriptorIndex, config_index);
	    exit(1);
	  }
#endif	  
	  // note that no *CCMcoeffs[config_index] to find in sum
	  // because it's the lhs.
	  
	  sum+=
	    hh[term].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
            indices[1]=4;
	    indices[2]=config_index;
            indices[3]=0;
	    indices[4]=0;
            indices[5]=0;
	    dummy=hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
	  }
	  
	}
      }
    }
  }
  free(configuration);
  return sum;
}

double term5(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 5: 

     F(k) . F(m) . S(+)(k) . S(+)(m).
     
  */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  //  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  // double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  int (*configuration)[5];

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; 
	for(m=1;m<5;m++) conf2[1][m]=configuration[k2][m];
	
	// loop over nearest neighbours
	
        index = conf1[1][4];
        are_they_neighbours=
          test_k_to_m_neighbour(hh,term, conf1[1], conf2[1]);
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=0;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      counter++;
	      tempArray[counter]=l;
	    }
	  }	  
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=1;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) {
	      if (spin[j]) {
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++)
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    findDescriptorOfCluster(conf1_len,
					  conf1, descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    if (descriptorIndex1 <= 0) continue;

	    findDescriptorOfCluster(conf2_len,
					  conf2, descriptor);
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
            if (descriptorIndex2 <= 0) continue;
	    // now find the multiplying cofficient.
	    
	    // multiplier=1.0;
	    
	    // for(i=1;i<conf1_len;i++)
	    //   multiplier=multiplier*i;
	    // multiplier=multiplier*conf1_len;
	    
	    // configuration 1
	    // (length-1)! -- permutation factor because we sum over k.
	    // length      -- factor from F(k).
	    
	    // for(i=1;i<conf2_len;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf2_len;
	    
	    // configuration 2
	    // (length-1)! -- permutation factor because we sum over m.
	    // length      -- factor from F(m).
	    
	    // rhs this time!
	    
	    // if descriptorIndex is 0, it means no matching cluster
	    // from the `fundamental' set could be found and so 
	    // this case is ignored -- i.e., not saved to disk.
	    
	    // if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	    //  sum+=CCMcoeffs[descriptorIndex1]
		//  *CCMcoeffs[descriptorIndex2]
  		//  *hh[term].coefficient[index][are_they_neighbours];
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=5;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
  dummy=hh[term].coefficient[index][are_they_neighbours];
                collect_term(indices,dummy,h_id);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term6(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  
  /* TERM 6: 
     
     S(+)(m) . S(z)(k)
     
  */
  
  int neighbour;
  double multiplier;
  int index;
  double sum;
  int new_index;
  
  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  sum=0.0;
  
  if (length != 1)
    return 0.0;
  
  index=string[1][4];
        
  for(neighbour=1;neighbour<=hh[term].m_to_k_number[index];
      neighbour++) {
    
    // nearest neighbour is at x,y,z, cell put information
    // into configuration and exit.
    
    /* m ---> k */
    
    new_index=hh[term].m_to_k[index][neighbour][3];
    
    // cout << "new Index = " << new_index << endl;
    
    multiplier=1.0;
    multiplier=multiplier*(-0.5*UnitCell.s2[new_index]);
    
    sum+=multiplier*hh[term].coefficient[index][neighbour];
    
    if (Control.file) {
      int indices[7]={0};
      double dummy;
      indices[0]=config_index;
      indices[1]=6;
      indices[2]=0; // !Special Case! No CCM ket-state coeffs here!
      indices[3]=0; // This is checked for in solve.C
      indices[4]=0;
      indices[5]=0;
      dummy=multiplier*hh[term].coefficient[index][neighbour];
       collect_term(indices,dummy,h_id);
    }
    
  }
  return sum;
}

// s^z s^ + all s

double term7(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 7: 
     
     F(k,l) . S(+)(k) . S(+)(m).
     
  */
  
  int i, j, l;
  int k1;
  int k2;
  // int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int tempConfig[MAX_LSUBM+2];
  int counter;
  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf=         (int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) {
      configuration[i][j]=string[i][j];
      conf[i][j]=0;
    }
  
  multiplier=1.0;
  
  // for(i=1;i<length-1;i++)
  //   multiplier=multiplier*i;
  // multiplier=multiplier*(length-1);
  
  // configuration 1
  // (length-2)!     -- permutation factor because we sum over k1 and k2.
  // (length-1)      -- factor from F(). Note that s_k2^+ `captures' one
  //                    of the spin flips but does not contribute to F().
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {  // ToDo: better k1 < k2 and factor 2 ??? Joerg
	
	for(l=1;l<5;l++) { conf[1][l]=configuration[k1][l]; }
	for(l=1;l<5;l++) { tempConfig[l]=configuration[k2][l]; }
	
	// loop over nearest neighbours
	
        index = conf[1][4];
        are_they_neighbours=
          test_k_to_m_neighbour(hh,term, conf[1], tempConfig);
	
	// conf[1][] and tempConfig[] need to be nearest-neighbours 
	
	// now load the rest into conf
	
	counter=1;
	for(i=1;i<=length;i++) {
	  if ((i != k2) && (i != k1)) {
	    counter++;
	    for(l=1;l<5;l++)
	      conf[counter][l]=configuration[i][l];
	  }
	}
	
	if (counter != (length-1)) {
	  printf("TERM 7: Error counter != (length-1)...\nExiting\n\n");
	  exit(1);
	}
	
	if (are_they_neighbours) {
	  
	  findDescriptorOfCluster(length-1,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  // if (descriptorIndex > 0)
	  //  sum+=multiplier*CCMcoeffs[descriptorIndex]*
	  //    hh[term].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=7;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*hh[term].coefficient[index][are_they_neighbours];
              collect_term(indices,dummy,h_id);
	    }
	  }
	  
	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

double term8(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 8: 
     
     -2 * F(m)(l) . S(z)(k) . S(z)(m)
     
     Note that this term generally cancels with term 17.......

     */

  int i, j, l;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  // double multiplier;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  int minX, minY, minZ;
  int maxX, maxY, maxZ;

  int x,y,z,cell;
  int index;
  
  int numberFlips, maxNumberFlips;
  int nn[5];

  int (*configuration)[5];

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);
  // Oct05: configs with length+1 does not generate terms, so we skip it
  if (length+1>Params.n) return 0;

  configuration=(int (*)[5])malloc((MAX_LSUBM+4)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  int maxPointX, maxPointY, maxPointZ;

  maxPointX=maxPointY=maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1])) maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2])) maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3])) maxPointZ=abs(string[j][3]);
  }

  if (Params.dim == 1) {
    minX=radius;           minY=0;                minZ=0;
    maxX=radius+maxPointX; maxY=0;                maxZ=0;
  } else if (Params.dim == 2) {
    minX=radius;           minY=radius;           minZ=0;
    maxX=radius+maxPointX; maxY=radius+maxPointY; maxZ=0;
  } else if (Params.dim == 3) {
    minX=radius;           minY=radius;           minZ=radius;
    maxX=radius+maxPointX; maxY=radius+maxPointY; maxZ=radius+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d)"
      " not recognised...\nExiting||\n\n", Params.dim);
    exit(1);
  }

  // multiplier=1.0;
  //  for(i=1;i<=length;i++)
  //  multiplier=multiplier*i;
  // multiplier=multiplier*(length+1);

  // length!               -- permutation factor
  // (length+1)            -- factor from F(m)

  for(x=-minX;x<=maxX;x++) {
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {
	  
	  // order=length+1;	            // this is the `m'    
	  
	  configuration[length+1][1]=x;
	  configuration[length+1][2]=y;
	  configuration[length+1][3]=z;
	  configuration[length+1][4]=cell;
	  
	  index=cell;
	  numberFlips=1;
	  
	  // check that we don't have too many spins at site `m'.
	  
	  for(l=1;l<=length;l++)
	    if (configuration[length+1][1]==configuration[l][1] &&
		configuration[length+1][2]==configuration[l][2] &&
		configuration[length+1][3]==configuration[l][3] &&
		configuration[length+1][4]==configuration[l][4])
	      numberFlips++;
	  
	  maxNumberFlips=UnitCell.s2[configuration[length+1][4]];
	  
	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	  
	  if (numberFlips <= maxNumberFlips) {
	    
	    for(j=1;j<=hh[term].m_to_k_number[index];j++) {
	      
	      // find the s_k^z, and so use
	      
	      /* m --> k */
	      
	      nn[1]=configuration[length+1][1]
		+hh[term].m_to_k[index][j][0];
	      nn[2]=configuration[length+1][2]
		+hh[term].m_to_k[index][j][1];
	      nn[3]=configuration[length+1][3]
		+hh[term].m_to_k[index][j][2];
	      nn[4]=hh[term].m_to_k[index][j][3];
	      
	      temp_multiplier=-2*0.5*UnitCell.s2[configuration[length+1][4]]*
		0.5*UnitCell.s2[nn[4]];
	      
	      findDescriptorOfCluster(length+1,
					    configuration,
					    descriptor);
	      
	      descriptorIndex=findDescriptorIndex(descriptor,
						  NumberConfigurations);
	      
	      // if (descriptorIndex > 0)
		// sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
		//  hh[term].coefficient[index][j];
	      
	      if (Control.file) {
		if (descriptorIndex > 0){
		  int indices[7]={0};
		  double dummy;
		  indices[0]=config_index;
		  indices[1]=8;
		  indices[2]=descriptorIndex;
		  indices[3]=0;
		  indices[4]=0;
		  indices[5]=0;
		  dummy=temp_multiplier*
		    hh[term].coefficient[index][j];
                  collect_term(indices,dummy,h_id);
		}
	      }
	      
	    }
	  }
	}
      }
    }    
  }
  return sum;  
}

double term9(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 9:
     
     -2 * H(k)(m)(l) . S(+)(k) . S(z)(m)
     
  */

  int i,j,k,l;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  multiplier=1.0;

  //  for(i=1;i<length;i++)
  //  multiplier=multiplier*i; // permutation factor
  // multiplier=multiplier*(length+1)*length; // factor from H()
  
  // (length-1)!        -- permutation factor as we sum over k
  // (length+1)*length  -- internal factor from H(). Note that 
  //                       summation over n.n. gives m thus 
  //                       length ->length+1.

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    
    for(neighbour=1;neighbour<=hh[term].k_to_m_number[index];
	neighbour++) {
      
      // nearest neighbour is at x,y,z, cell put information
      // into configuration and exit.
	
      /* k ---> m */
	
      nn[1]=configuration[k][1]+
	hh[term].k_to_m[index][neighbour][0];
      nn[2]=configuration[k][2]+
	hh[term].k_to_m[index][neighbour][1];
      nn[3]=configuration[k][3]+
	hh[term].k_to_m[index][neighbour][2];
      nn[4]=hh[term].k_to_m[index][neighbour][3];
	
      numberFlips=1;
	
      for(l=1;l<=length;l++)
	if (nn[1]==configuration[l][1] &&
	    nn[2]==configuration[l][2] &&
	    nn[3]==configuration[l][3] &&
	    nn[4]==configuration[l][4])
	  numberFlips++;
	
      maxNumberFlips=UnitCell.s2[nn[4]];
	
      // the number of spin flips that you can have at a given
      // lattice site is depedent on the quantum spin number
      // of the site in the UNIT CELL index `length'
	
      if (numberFlips <= maxNumberFlips) {
	  
	configuration[length+1][1]=nn[1];
	configuration[length+1][2]=nn[2];
	configuration[length+1][3]=nn[3];
	configuration[length+1][4]=nn[4];
	  
	findDescriptorOfCluster((length+1),configuration,descriptor);
	descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
	  
	// -2 * -0.5*UnitCell.s2[] from coefficient factors in the term 8

	temp_multiplier=2*multiplier*0.5*UnitCell.s2[nn[4]];
	  
	// if (descriptorIndex > 0)
	//  sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
	//    hh[term].coefficient[index][neighbour];;	
	  
	if (Control.file) {
	  if (descriptorIndex > 0) {
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=9;
	    indices[2]=descriptorIndex;
	    indices[3]=0;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=temp_multiplier*hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	  }
	}
	  
      }
    }
  }
  
  free(configuration);
  return sum;
}  

double term10(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 10: 
     
     - F(m) . F(m) . S(+)(m) . S(z)(k)

     */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // F(m)
    
    for(m=1;m<5;m++)
      conf2[1][m]=configuration[k][m]; // F(m)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(m) 1st one
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // F(m) 2nd one
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];
      
      for(neighbour=1;neighbour<=
	    hh[term].m_to_k_number[index];neighbour++) {
	
	// find s_k^z and so use
	
	/* m ---> k */
	
	nn[1]=conf1[1][1]+
	  hh[term].m_to_k[index][neighbour][0];
	nn[2]=conf1[1][2]+
	  hh[term].m_to_k[index][neighbour][1];
	nn[3]=conf1[1][3]+
	  hh[term].m_to_k[index][neighbour][2];
	nn[4]=hh[term].m_to_k[index][neighbour][3];
	
	findDescriptorOfCluster(conf1_len,
				      conf1,
				      descriptor);
	descriptorIndex1=findDescriptorIndex(descriptor,
					     NumberConfigurations);
	
                if (descriptorIndex1 <= 0) continue;
	findDescriptorOfCluster(conf2_len,
				      conf2,
				      descriptor);
	
	descriptorIndex2=findDescriptorIndex(descriptor,
					     NumberConfigurations);
	
                if (descriptorIndex2 <= 0) continue;
	multiplier=1.0;
	// for(i=1;i<=conf1_len-1;i++)
	//   multiplier=multiplier*i; 
	// multiplier=multiplier*conf1_len;
	
	// F(m)
	// (length-1)!       -- permutation factor from F(m)
	// length            -- internal factor in F(m)
	
	// for(i=1;i<=conf2_len-1;i++)
	//  multiplier=multiplier*i;
	// multiplier=multiplier*conf2_len;
	
	// F(m)
	// (length-1)!       -- permutation factor from F(m)
	// length            -- internal factor in F(m)
	
	multiplier=multiplier*0.5*UnitCell.s2[nn[4]];
	
	// coeff from \tilde s^- \tilde s^z
	
	// if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	//  sum+=multiplier
	//    *CCMcoeffs[descriptorIndex1]
	//    *CCMcoeffs[descriptorIndex2]
	//    *hh[term].coefficient[index][neighbour];
	
	if (Control.file) {
	  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=10;
	    indices[2]=descriptorIndex1;
	    indices[3]=descriptorIndex2;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=multiplier*hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term11(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 11: 
     
     -2 * H(k)(m) . F(m) . S(+)(k) . S(+)(m)
     
    */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  //  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // H(k,m) <-- k here
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k2][m]; // F(m)
	
	// loop over nearest neighbours
	
        index = conf1[1][4];
        are_they_neighbours=
          test_k_to_m_neighbour(hh,term, conf1[1], conf2[1]);
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  for(m=1;m<5;m++)
	    conf1[2][m]=configuration[k2][m];
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=2;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		// H(k,m)
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		// F(m)
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    findDescriptorOfCluster(conf1_len,
					  conf1,
					  descriptor);
	    
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex1 <= 0) continue;
	    findDescriptorOfCluster(conf2_len,
					  conf2,
					  descriptor);
	    
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex2 <= 0) continue;
	    multiplier=1.0;
	    // for(i=1;i<=(conf1_len-2);i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf1_len*
	    //  (conf1_len-1);
	    
	    // H(k,m)             
	    // (length-2)!        -- because we sum over k and m
	    // length*(length-1)  -- factor due to H().
	    
	    // for(i=1;i<=conf2_len-1;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf2_len;
	    
	    // F(m)             
	    // (length-1)!        -- because we sum over k and m
	    // length             -- factor due to F().
	    
	    multiplier=-2.0;
	    
	    // coeff from \tilde s^z \tilde s^-
	    
	    // if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	    //  sum+=multiplier*CCMcoeffs[descriptorIndex1]
		//*CCMcoeffs[descriptorIndex2]
		//*hh[term].coefficient[index][are_they_neighbours];
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=11;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term12(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 12:
     
     -2 * F(k) . F(m) . S(+)(k) . S(z)(m)
     
  */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // F(k)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=0;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(k)
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // F(m)
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];
      
      for(neighbour=1;neighbour<=hh[term].k_to_m_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=conf1[1][1]+hh[term].k_to_m[index][neighbour][0];
	nn[2]=conf1[1][2]+hh[term].k_to_m[index][neighbour][1];
	nn[3]=conf1[1][3]+hh[term].k_to_m[index][neighbour][2];
	nn[4]=hh[term].k_to_m[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_len;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_len+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  findDescriptorOfCluster(conf1_len,
					conf1,
					descriptor);
	  descriptorIndex1=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	  
                if (descriptorIndex1 <= 0) continue;
	  findDescriptorOfCluster(temp_number_conf2_sites,
					conf2,
					descriptor);
	  
	  descriptorIndex2=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	  
                if (descriptorIndex2 <= 0) continue;
	  multiplier=1.0;
	  // for(i=1;i<=conf1_len-1;i++)
	  //  multiplier=multiplier*i; 
	  // multiplier=multiplier*conf1_len;
	  
	  // F(k)
	  // (length-1)!       -- because we sum over k
	  // length            -- internal factor in F(k)
	  
	  // for(i=1;i<=temp_number_conf2_sites-1;i++)
	  //  multiplier=multiplier*i;
	  // multiplier=multiplier*temp_number_conf2_sites; 
	  
	  // F(m)
	  // (length-1)!       -- permutation factor from F(m)
	  // length            -- internal factor in F(m) because `m' 
	  //                      is `virtual'. (Note that summation over 
	  //                      n.n. gives m thus length ->length+1.)
	  
	  multiplier=2*multiplier*0.5*UnitCell.s2[nn[4]];
	  
	  // coeff from \tilde s^- \tilde s^z
	  
	  //if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	    //sum+=multiplier
	    //  *CCMcoeffs[descriptorIndex1]
	    //  *CCMcoeffs[descriptorIndex2]*
	    //  hh[term].coefficient[index][neighbour];
	  
	  if (Control.file) {
	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=12;
	      indices[2]=descriptorIndex1;
	      indices[3]=descriptorIndex2;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  


double term13(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 13: 
     
     - F(k) . F(m) . F(m) . S(+)(k) . S(+)(m)
   
    */

  int spin[MAX_LSUBM+1];
  int counter;
  int mm,nnn,temp1,temp2,temp3;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  // int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; // F(k)
	for(m=1;m<5;m++) conf2[1][m]=configuration[k2][m]; // F(m)
	for(m=1;m<5;m++) conf3[1][m]=configuration[k2][m]; // F(m)
	
	// loop over nearest neighbours
	
        index = conf1[1][4];
        are_they_neighbours=
          test_k_to_m_neighbour(hh,term, conf1[1], conf2[1]);
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
          // Joerg: I dont know, why splitting here
          //        but it makes program slow for counter > 10 ???
          //   trimary_max[i] = 3^i, 30*ln3=47.55*ln2 48bit=6*8bit
          // ToDo: simplify spin[i]=000 001 002 010 011 012 020 021 ...
	  if (counter > 30) {
	    printf("TERM 13: counter > 30...\nExiting!!\n\n");
	    exit(1);
	  }
          temp = counter; // split into summands not greater than 10 each
	  temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
	  temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
	  temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	  
          // trimary_max[i] = 3^i  , try all 3^counter codes, counter<SUBn
	  for(l=0;l<trimary_max[temp1];l++) {
	    for(mm=0;mm<trimary_max[temp2];mm++) {
	      for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		
                // decode counter to spin[i] trimary code
		trimary(l,mm,nnn,counter,spin);
                // now we have next trimary code 000 -> 001 -> 002 -> 010 ...

		conf1_len=1;
		conf2_len=1;
		conf3_len=1;
		
		for (j=1;j<=counter;j++) { 
		  if (spin[j]==0) {         // F(k)
		    conf1_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf1[conf1_len][m]=configuration[temp][m];
		  } else if (spin[j]==1) {  // F(m)
		    conf2_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf2[conf2_len][m]=configuration[temp][m];
		  } else {                   // F(m)
		    conf3_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf3[conf3_len][m]=configuration[temp][m];
		  }
		}
		
		findDescriptorOfCluster(conf1_len,
					      conf1, descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		findDescriptorOfCluster(conf2_len,
					      conf2, descriptor);
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex2 <= 0) continue;
		findDescriptorOfCluster(conf3_len,
					      conf3, descriptor);	      
		descriptorIndex3=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex3 <= 0) continue;
		multiplier=1.0;
		
		// for(i=1;i<=(conf1_len-1);i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf1_len;
		
		// for(i=1;i<=(conf2_len-1);i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf2_len;
		
		// for(i=1;i<=(conf3_len-1);i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf3_len;
		
		// in each case:
		// (length-1)!      -- we sum over k1 and k2.
		// length           -- internal factor in F().
		
		multiplier=-1.0;
		
		// coeff from \tilde s^z \tilde s^-
		
		//if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)
		//    && (descriptorIndex3 > 0))
		//  sum+=multiplier
		//    *CCMcoeffs[descriptorIndex1]
		//    *CCMcoeffs[descriptorIndex2]
		//    *CCMcoeffs[descriptorIndex3]
		//    *hh[term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) 
		      && (descriptorIndex3 > 0)){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=13;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=descriptorIndex3;
		    indices[5]=0;
		    dummy=multiplier*hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
		  }
		}
		
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}

// s^+ s^z 

double term14(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  
  
  /* TERM 14: 
     
     S(+)(k) . S(z)(m)
     
  */
  
  int neighbour;
  double multiplier;
  int index;
  int new_index;
  double sum;

  sum=0.0;

  if (length != 1)
    return 0.0;
  
  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  index=string[1][4];
  
    for(neighbour=1;neighbour<=hh[term].k_to_m_number[index];
	neighbour++) {
      
      // nearest neighbour is at x,y,z, cell put information
      // into configuration and exit.
      
      /* k ---> m */
      
      new_index=hh[term].k_to_m[index][neighbour][3];

      multiplier=1.0;
      multiplier=multiplier*(-1*0.5*UnitCell.s2[new_index]);
      
      sum+=multiplier*hh[term].coefficient[index][neighbour];
      
      if (Control.file) {
	int indices[7]={0};
	double dummy;
	indices[0]=config_index;
	indices[1]=14;
	indices[2]=0;
	indices[3]=0;
	indices[4]=0;
	indices[5]=0;
	dummy=multiplier*hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
      }
      
  }
  return sum;
}

double term15(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 15: 
     
     F(m) . S(+)(m) . S(+)(k).
     
  */


  int i, j, l;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int tempConfig[MAX_LSUBM+2];
  int counter;
  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf=         (int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) {
      configuration[i][j]=string[i][j];
      conf[i][j]=0;
    }
  
  multiplier=1.0;
  
  //  for(i=1;i<length-1;i++)
  //  multiplier=multiplier*i;
  // multiplier=multiplier*(length-1);
  
  // configuration 1
  // (length-2)!     -- permutation factor because we sum over k1 and k2.
  // (length-1)      -- factor from F(). Note that s_k2^+ `captures' one
  //                    of the spin flips but does not contribute to F().
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {
	
	for(l=1;l<5;l++) { conf[1][l]=configuration[k1][l]; }    // F(m) 
	for(l=1;l<5;l++) { tempConfig[l]=configuration[k2][l]; } // s(k)^+
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf[1][4];      
	
	for(j=1;j<=hh[term].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  /* m ---> k */
	  
	  nn[1]=conf[1][1]+hh[term].m_to_k[index][j][0];
	  nn[2]=conf[1][2]+hh[term].m_to_k[index][j][1];
	  nn[3]=conf[1][3]+hh[term].m_to_k[index][j][2];
	  nn[4]=           hh[term].m_to_k[index][j][3];
	  
	    if(tempConfig[1] == nn[1] &&
	       tempConfig[2] == nn[2] &&
	       tempConfig[3] == nn[3] &&
	       tempConfig[4] == nn[4])
	      are_they_neighbours=j;      
	  
	  if (are_they_neighbours) break;
	}
	
	// now load the rest into conf
	
	counter=1;
	for(i=1;i<=length;i++) {
	  if ((i != k2) && (i != k1)) {
	    counter++;
	    for(l=1;l<5;l++)
	      conf[counter][l]=configuration[i][l];
	  }
	}
	
	if (counter != (length-1)) {
	  printf("TERM 15: Error counter != (length-1)...\nExiting\n\n");
	  exit(1);
	}
	
	if (are_they_neighbours) {
	  
	  findDescriptorOfCluster(length-1,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
//	  if (descriptorIndex > 0)
//	    sum+=multiplier*CCMcoeffs[descriptorIndex]*
//	      hh[term].coefficient[index][are_they_neighbours];	
	  
	  if (Control.file) {
	    if (descriptorIndex > 0){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=15;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
	    }
	  }
	  
	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

// s^- s^z


double term16(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 16:
     
     -2 * H(m)(k)(l) . S(+)(m) . S(z)(k)
     
  */

  int i,j,k,l;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  multiplier=1.0;
  //   for(i=1;i<length;i++)
  //  multiplier=multiplier*i; 
  //   multiplier=multiplier*(length+1)*length; 
  
  // (length-1)!        -- permutation factor as we sum over k
  // (length+1)*length  -- internal factor from H(). Note that 
  //                       summation over n.n. gives m thus 
  //                       length -> length+1.

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    
    for(neighbour=1;neighbour<=hh[term].m_to_k_number[index];
	neighbour++) {
      
      // nearest neighbour is at x,y,z, cell put information
      // into configuration and exit.
      
      /* m ---> k */
      
      nn[1]=configuration[k][1]+
	hh[term].m_to_k[index][neighbour][0];
      nn[2]=configuration[k][2]+
	hh[term].m_to_k[index][neighbour][1];
      nn[3]=configuration[k][3]+
	hh[term].m_to_k[index][neighbour][2];
      nn[4]=hh[term].m_to_k[index][neighbour][3];

      
      numberFlips=1;
      
      for(l=1;l<=length;l++)
	if (nn[1]==configuration[l][1] &&
	    nn[2]==configuration[l][2] &&
	    nn[3]==configuration[l][3] &&
	    nn[4]==configuration[l][4])
	  numberFlips++;
      
      maxNumberFlips=UnitCell.s2[nn[4]];
      
      // the number of spin flips that you can have at a given
      // lattice site is depedent on the quantum spin number
      // of the site in the UNIT CELL index `length'
      
      if (numberFlips <= maxNumberFlips) {
	
	configuration[length+1][1]=nn[1];
	configuration[length+1][2]=nn[2];
	configuration[length+1][3]=nn[3];
	configuration[length+1][4]=nn[4]; // H(m,k)
	
	findDescriptorOfCluster(length+1,configuration,descriptor);
	descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
	
	temp_multiplier=multiplier*2*0.5*UnitCell.s2[nn[4]];
	
	// coeff from \tilde s^- \tilde s^z
	
//	if (descriptorIndex > 0)
//	  sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
//	    hh[term].coefficient[index][neighbour];	
	
	if (Control.file) {
	  if (descriptorIndex > 0){
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=16;
	    indices[2]=descriptorIndex;
	    indices[3]=0;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=temp_multiplier*hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	  }
	}
	
      }
    }
  }
  free(configuration);
  return sum;
}  

double term17(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 17: 
     
     -2 * F(m)(l) * S(z)(k) * S(z)(m)

     */

  int i, j, l;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  // double multiplier;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  int minX, minY, minZ;
  int maxX, maxY, maxZ;
  int x,y,z,cell;
  int index;

  int numberFlips, maxNumberFlips;
  int nn[5];

  int (*configuration)[5];

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);
  // Oct05: configs with length+1 does not generate terms, so we skip it
  if (length+1>Params.n) return 0;

  configuration=(int (*)[5])malloc((MAX_LSUBM+4)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  int maxPointX;
  int maxPointY;
  int maxPointZ;

  maxPointX=0;
  maxPointY=0;
  maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1]))
      maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2]))
      maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3]))
      maxPointZ=abs(string[j][3]);
  }

  if (Params.dim == 1) {
    minX=radius;
    minY=0;
    minZ=0;
    maxX=radius+maxPointX;
    maxY=0;
    maxZ=0;
  } else if (Params.dim == 2) {
    minX=radius;
    minY=radius;
    minZ=0;
    maxX=radius+maxPointX;
    maxY=radius+maxPointY;
    maxZ=0;
  } else if (Params.dim == 3) {
    minX=radius;
    minY=radius;
    minZ=radius;
    maxX=radius+maxPointX;
    maxY=radius+maxPointY;
    maxZ=radius+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d"
      ") not recognised...\nExiting||\n\n",Params.dim);
    exit(1);
  }

  // multiplier=1.0;
  // for(i=1;i<=length;i++)
  //  multiplier=multiplier*i;
  // multiplier=multiplier*(length+1);

  // length!               -- permutation factor
  // (length+1)            -- factor from F(k)

  for(x=-minX;x<=maxX;x++) {
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {

	  // order=length+1;	            // this is the `k'    
	  
	  configuration[length+1][1]=x;
	  configuration[length+1][2]=y;
	  configuration[length+1][3]=z;
	  configuration[length+1][4]=cell;
	  
	  index=cell;
	  numberFlips=1;
	  
	  // check that we don't have too many spins at site `k'.

	  for(l=1;l<=length;l++)
	    if (configuration[length+1][1]==configuration[l][1] &&
		configuration[length+1][2]==configuration[l][2] &&
		configuration[length+1][3]==configuration[l][3] &&
		configuration[length+1][4]==configuration[l][4])
	      numberFlips++;
	  
	  maxNumberFlips=UnitCell.s2[configuration[length+1][4]];
	  
	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	  
	  if (numberFlips <= maxNumberFlips) {
	    
	    for(j=1;j<=hh[term].k_to_m_number[index];j++) {
	      
	      // find the s_k^m, and so use
	      
	      /* k --> m */
	      
	      nn[1]=configuration[length+1][1]
		+hh[term].k_to_m[index][j][0];
	      nn[2]=configuration[length+1][2]
		+hh[term].k_to_m[index][j][1];
	      nn[3]=configuration[length+1][3]
		+hh[term].k_to_m[index][j][2];
	      nn[4]=hh[term].k_to_m[index][j][3];
	      
	      temp_multiplier=-2*0.5*UnitCell.s2[configuration[length+1][4]]*
		0.5*UnitCell.s2[nn[4]];
	      
	      findDescriptorOfCluster(length+1,
					    configuration,
					    descriptor);
	      
	      descriptorIndex=findDescriptorIndex(descriptor,
						  NumberConfigurations);
	      
//	      if (descriptorIndex > 0)
//		sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
//		  hh[term].coefficient[index][j];
	      
	      if (Control.file) {
		if (descriptorIndex > 0){
		  int indices[7]={0};
		  double dummy;
		  indices[0]=config_index;
		  indices[1]=17;
		  indices[2]=descriptorIndex;
		  indices[3]=0;
		  indices[4]=0;
		  indices[5]=0;
		  dummy=temp_multiplier*
		    hh[term].coefficient[index][j];
            collect_term(indices,dummy,h_id);
		}
	      }
	      
	    }
	  }
	}
      }
    }
    
  }
  return sum;  
}

double term18(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 18:
     
     -2 * F(m) . F(k) . S(+)(m) . S(z)(k)
     
  */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // F(m)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    counter--;

    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=0;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(m)
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // F(k)
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];
      
      
      for(neighbour=1;neighbour<=hh[term].m_to_k_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* m ---> k */
	
	nn[1]=conf1[1][1]+hh[term].m_to_k[index][neighbour][0];
	nn[2]=conf1[1][2]+hh[term].m_to_k[index][neighbour][1];
	nn[3]=conf1[1][3]+hh[term].m_to_k[index][neighbour][2];
	nn[4]=hh[term].m_to_k[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_len;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_len+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  findDescriptorOfCluster(conf1_len,
					conf1,
					descriptor);
	  descriptorIndex1=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	  
                if (descriptorIndex1 <= 0) continue;
	  findDescriptorOfCluster(temp_number_conf2_sites,
					conf2,
					descriptor);
	  
	  descriptorIndex2=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	  
                if (descriptorIndex2 <= 0) continue;
	  multiplier=1.0;
	  // for(i=1;i<conf1_len;i++)
	  //  multiplier=multiplier*i; 
	  // multiplier=multiplier*conf1_len;
	  
	  // F(k)
	  // (length-1)!       -- because we sum over k
	  // length            -- internal factor in F(k)
	  
	  // for(i=1;i<temp_number_conf2_sites;i++)
	  //  multiplier=multiplier*i;
	  // multiplier=multiplier*temp_number_conf2_sites; 
	  
	  // F(m)
	  // (length-1)!       -- permutation factor from F(m)
	  // length            -- internal factor in F(m) because `m' 
	  //                      is `virtual'. (Note that summation over 
	  //                      n.n. gives m thus length ->length+1.)
	  
	  multiplier=2*multiplier*0.5*UnitCell.s2[nn[4]];
	  
	  // coeff from \tilde s^z \tilde s^-
	  
//	  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//	    sum+=multiplier
//	      *CCMcoeffs[descriptorIndex1]
//	      *CCMcoeffs[descriptorIndex2]*
//	      hh[term].coefficient[index][neighbour];
	  
	  if (Control.file) {
	    if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=18;
	      indices[2]=descriptorIndex1;
	      indices[3]=descriptorIndex2;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	    }
	  }
	  
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term19(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 19: 
     
     -2 * H(k)(m) . F(k) . S(+)(k) . S(+)(m)
     
    */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // H(m,k)
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k2][m]; // F(k)
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      
	
	for(j=1;j<=hh[term].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* m ---> k */
	  
	  nn[1]=conf1[1][1]+hh[term].m_to_k[index][j][0];
	  nn[2]=conf1[1][2]+hh[term].m_to_k[index][j][1];
	  nn[3]=conf1[1][3]+hh[term].m_to_k[index][j][2];
	  nn[4]=hh[term].m_to_k[index][j][3];
	  
	  if(conf2[1][1] == nn[1] &&
	     conf2[1][2] == nn[2] &&
	     conf2[1][3] == nn[3] &&
	     conf2[1][4] == nn[4])
	    are_they_neighbours=j;
      
	  if (are_they_neighbours) break;
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 

	if (are_they_neighbours) {

	  for(m=1;m<5;m++)
	    conf1[2][m]=configuration[k2][m];
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=2;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		// H
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		// H
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    findDescriptorOfCluster(conf1_len,
					  conf1,
					  descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex1 <= 0) continue;
	    findDescriptorOfCluster(conf2_len,
					  conf2,
					  descriptor);
	    
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex2 <= 0) continue;
	    multiplier=1.0;
	    // for(i=1;i<=(conf1_len-2);i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf1_len*
	    //  (conf1_len-1);
	    
	    // H(m,k)             
	    // (length-2)!        -- because we sum over k and m
	    // length*(length-1)  -- factor due to H().
	    
	    // for(i=1;i<=conf2_len-1;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf2_len;
	    
	    // F(m)             
	    // (length-1)!        -- because we sum over k and m
	    // length             -- factor due to F().
	    
	    multiplier=-2.0;
	    
	    // from \tilde s^- \tilde s^z
	    
//	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//	      sum+=multiplier
//		*CCMcoeffs[descriptorIndex1]
//		*CCMcoeffs[descriptorIndex2]*
//		hh[term].coefficient[index][are_they_neighbours];
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=19;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term20(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 20: 
     
     - F(m) . F(m) . S(+)(m)
     
  */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // F(k)
    
    for(m=1;m<5;m++)
      conf2[1][m]=configuration[k][m]; // F(k)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(k) 1st one
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // F(k) 2nd one
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];
      
      for(neighbour=1;neighbour<=
	    hh[term].k_to_m_number[index];neighbour++) {
	
	// find s_m^z and so use
	
	/* k ---> m */
	
	nn[1]=conf1[1][1]+hh[term].k_to_m[index][neighbour][0];
	nn[2]=conf1[1][2]+hh[term].k_to_m[index][neighbour][1];
	nn[3]=conf1[1][3]+hh[term].k_to_m[index][neighbour][2];
	nn[4]=hh[term].k_to_m[index][neighbour][3];
	
	findDescriptorOfCluster(conf1_len,
				      conf1,
				      descriptor);
	descriptorIndex1=findDescriptorIndex(descriptor,
					     NumberConfigurations);
	
                if (descriptorIndex1 <= 0) continue;
	findDescriptorOfCluster(conf2_len,
				      conf2,
				      descriptor);
	
	descriptorIndex2=findDescriptorIndex(descriptor,
					     NumberConfigurations);
	
                if (descriptorIndex2 <= 0) continue;
	multiplier=1.0;
	// for(i=1;i<=conf1_len-1;i++)
	//  multiplier=multiplier*i; 
	// multiplier=multiplier*conf1_len;
	
	// F(k)
	// (length-1)!       -- permutation factor from F(k)
	// length            -- internal factor in F(k)
	
	// for(i=1;i<=conf2_len-1;i++)
	//  multiplier=multiplier*i;
	// multiplier=multiplier*conf2_len;
	
	// F(k)
	// (length-1)!       -- permutation factor from F(k)
	// length            -- internal factor in F(k)
	
	multiplier=multiplier*0.5*UnitCell.s2[nn[4]];
	
	// coeff from \tilde s^- \tilde s^z
	
//	if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//	  sum+=multiplier
//	    *CCMcoeffs[descriptorIndex1]
//	    *CCMcoeffs[descriptorIndex2]*
//	    hh[term].coefficient[index][neighbour];
	
	if (Control.file) {
	  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=20;
	    indices[2]=descriptorIndex1;
	    indices[3]=descriptorIndex2;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=multiplier*hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum; 
}


double term21(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 21: 
     
     - F(k) . F(k) . F(m) . S(+)(k) . S(+)(m)
   
    */

  int spin[MAX_LSUBM+1];
  int mm,nnn,temp1,temp2,temp3;
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // F(m)
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k2][m]; // F(k)

	for(m=1;m<5;m++) 
	  conf3[1][m]=configuration[k2][m]; // F(k)
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      
	
	for(j=1;j<=hh[term].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* m ---> k */
	  
	  nn[1]=conf1[1][1]+hh[term].m_to_k[index][j][0];
	  nn[2]=conf1[1][2]+hh[term].m_to_k[index][j][1];
	  nn[3]=conf1[1][3]+hh[term].m_to_k[index][j][2];
	  nn[4]=hh[term].m_to_k[index][j][3];
	  
	  if(conf2[1][1] == nn[1] &&
	     conf2[1][2] == nn[2] &&
	     conf2[1][3] == nn[3] &&
	     conf2[1][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  if (counter > 30) {
	    printf("TERM 21: counter > 30...\nExiting!!\n\n");
	    exit(1);
	  }
          temp = counter; // split into summands not greater than 10 each
	  temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
	  temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
	  temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	  
	  for(l=0;l<trimary_max[temp1];l++) {
	    for(mm=0;mm<trimary_max[temp2];mm++) {
	      for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		
		trimary(l,mm,nnn,counter,spin);
		conf1_len=1;
		conf2_len=1;
		conf3_len=1;
		
		for (j=1;j<=counter;j++) { 
		  if (spin[j]==0) {           // F(m)
		    conf1_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf1[conf1_len][m]=configuration[temp][m];
		  } else if (spin[j]==1) {    // F(k)
		    conf2_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf2[conf2_len][m]=configuration[temp][m];
		  } else {                    // F(k)
		    conf3_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf3[conf3_len][m]=configuration[temp][m];
		  }
		}	      
		
		findDescriptorOfCluster(conf1_len,
					      conf1,
					      descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		findDescriptorOfCluster(conf2_len,
					      conf2,
					      descriptor);
		
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
		
                if (descriptorIndex2 <= 0) continue;
		findDescriptorOfCluster(conf3_len,
					      conf3,
					      descriptor);
		
		descriptorIndex3=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex3 <= 0) continue;
		multiplier=1.0;
		
		// for(i=1;i<=(conf1_len-1);i++)
		//     multiplier=multiplier*i;
		// multiplier=multiplier*conf1_len;
		
		// for(i=1;i<=(conf2_len-1);i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf2_len;
		
		// for(i=1;i<=(conf3_len-1);i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf3_len;
		
		  // in each case:
		// (length-1)!      -- we sum over k1 and k2.
		// length           -- internal factor in F().
		
		multiplier=-1.0;
		
		  // from \tilde s^- \tilde s^z
		
//		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)
//		    && (descriptorIndex3 > 0))
//		  sum+=multiplier
//		    *CCMcoeffs[descriptorIndex1]
//		    *CCMcoeffs[descriptorIndex2]
//		    *CCMcoeffs[descriptorIndex3]*
//		    hh[term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)
		      &&(descriptorIndex3 > 0)){
		      int indices[7]={0};
		      double dummy;
		      indices[0]=config_index;
		      indices[1]=21;
		      indices[2]=descriptorIndex1;
		      indices[3]=descriptorIndex2;
		      indices[4]=descriptorIndex3;
		      indices[5]=0;
		      dummy=multiplier*
			hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
		  }
		}
		
	      }
	    }
	  }
	}
      }
    }
  }

  
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}
 
double term22(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 22:
     
     - 2 * F(m) . S(+)(k) . S(z)(m)
     
  */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  // double multiplier;
  double temp_multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int counter;
  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
      
  for(i=1;i<=MAX_LSUBM+1;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;
      
  // multiplier=1.0;
  // for(i=1;i<length;i++)
  //  multiplier=multiplier*i;    // permutation factor
  // multiplier=multiplier*length; // factor from H()
  
  // (length-1)!        -- permutation factor as we sum over k
  // length             -- internal factor from F(). 

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    
    counter=0;
    for(l=1;l<=length;l++) {
      if (l != k) {
	counter++;
	for(m=1;m<5;m++)
	  conf[counter][m]=configuration[l][m];
      } 
    }
    for(neighbour=1;neighbour<=hh[term].k_to_m_number[index];
	neighbour++) {
      
      counter=length-1;
      
      // nearest neighbour is at x,y,z, cell put information
      // into configuration and exit.
      
      /* k ---> m */
      
      nn[1]=configuration[k][1]+
	hh[term].k_to_m[index][neighbour][0];
      nn[2]=configuration[k][2]+
	hh[term].k_to_m[index][neighbour][1];
      nn[3]=configuration[k][3]+
	hh[term].k_to_m[index][neighbour][2];
      nn[4]=hh[term].k_to_m[index][neighbour][3];
      
      numberFlips=1;
      
      for(l=1;l<=length;l++)
	if (nn[1]==configuration[l][1] &&
	    nn[2]==configuration[l][2] &&
	    nn[3]==configuration[l][3] &&
	    nn[4]==configuration[l][4])
	  numberFlips++;
      
      maxNumberFlips=UnitCell.s2[nn[4]];
      
      // the number of spin flips that you can have at a given
      // lattice site is depedent on the quantum spin number
      // of the site in the UNIT CELL index `length'
      
      if (numberFlips <= maxNumberFlips) {
	
	counter++;
	
	conf[counter][1]=nn[1];
	conf[counter][2]=nn[2];
	conf[counter][3]=nn[3];
	conf[counter][4]=nn[4];
	
	if (counter != length) {
	  printf("TERM 22: Error counter != length...\nExiting!!\n\n");
	  exit(1);
	}
	
	findDescriptorOfCluster(counter,conf,descriptor);
	descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
	
	temp_multiplier=2*0.5*UnitCell.s2[nn[4]];
	
	// coeff from \tilde s^+ \tilde s^-
	
//	if (descriptorIndex > 0)
//	  sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
//	    hh[term].coefficient[index][neighbour];	
	
	if (Control.file) {
	  if (descriptorIndex > 0){
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=22;
	    indices[2]=descriptorIndex;
	    indices[3]=0;
	    indices[4]=0;
            indices[5]=0;
	    dummy=temp_multiplier*hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	  }
	}
	
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}  

double term23(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  /* TERM 23: 

     - F(m) . F(m) . S(+)(k) . S(+)(m).
     
  */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int tempString[5];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // F(m)
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k1][m]; // F(m)
	
	for(m=1;m<5;m++) 
	  tempString[m]=configuration[k2][m]; // s(+)(k)
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      
	
	for(j=1;j<=hh[term].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* m ---> k */
	  
	  nn[1]=conf1[1][1]+hh[term].m_to_k[index][j][0];
	  nn[2]=conf1[1][2]+hh[term].m_to_k[index][j][1];
	  nn[3]=conf1[1][3]+hh[term].m_to_k[index][j][2];
	  nn[4]=hh[term].m_to_k[index][j][3];
	    
	  if(tempString[1] == nn[1] &&
	     tempString[2] == nn[2] &&
	     tempString[3] == nn[3] &&
	     tempString[4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}

	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=1;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    findDescriptorOfCluster(conf1_len,
					  conf1,
					  descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex1 <= 0) continue;
	    findDescriptorOfCluster(conf2_len,
					  conf2,
					  descriptor);
	    
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
	    
                if (descriptorIndex2 <= 0) continue;
	    // for(i=1;i<conf1_len;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf1_len;
	    
	    // configuration 1
	    // (length-1)! -- permutation factor because we sum over m.
	    // length      -- factor from F(m)
	    
	    // for(i=1;i<conf2_len;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf2_len;
	    
	    // configuration 2
	    // (length-1)! -- permutation factor because we sum over m.
	    // length      -- factor from F(m)
	    
	    multiplier=-1.0;
	    
	    // coeff from \tilde s^+ \tilde s^-
	    
//	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//	      sum+=multiplier
//		*CCMcoeffs[descriptorIndex1]
//		*CCMcoeffs[descriptorIndex2]*
//		hh[term].coefficient[index][are_they_neighbours];
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=23;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term24(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 24:
     
     -2 * F(k) . S(+)(m) . S(z)(k)
     
  */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  // double multiplier;
  double temp_multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int counter;
  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
      
  for(i=1;i<=MAX_LSUBM+1;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;
      
  // multiplier=1.0;
  // for(i=1;i<length;i++)
  //  multiplier=multiplier*i;    
  // multiplier=multiplier*length; 
  
  // (length-1)!        -- permutation factor as we sum over m
  // length             -- internal factor from F(). 

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    
    counter=0;
    for(l=1;l<=length;l++) {
      if (l != k) {
	counter++;
	for(m=1;m<5;m++)
	  conf[counter][m]=configuration[l][m];
      }
    }
    
    for(neighbour=1;neighbour<=hh[term].m_to_k_number[index];
	neighbour++) {
      
      counter=length-1;
      
      // nearest neighbour is at x,y,z, cell put information
      // into configuration and exit.
      
      /* m ---> k */
      
      nn[1]=configuration[k][1]+
	hh[term].m_to_k[index][neighbour][0];
      nn[2]=configuration[k][2]+
	hh[term].m_to_k[index][neighbour][1];
      nn[3]=configuration[k][3]+
	hh[term].m_to_k[index][neighbour][2];
      nn[4]=hh[term].m_to_k[index][neighbour][3];
      
      numberFlips=1;
      
      for(l=1;l<=length;l++)
	if (nn[1]==configuration[l][1] &&
	    nn[2]==configuration[l][2] &&
	    nn[3]==configuration[l][3] &&
	    nn[4]==configuration[l][4])
	  numberFlips++;
      
      maxNumberFlips=UnitCell.s2[nn[4]];
      
      // the number of spin flips that you can have at a given
      // lattice site is depedent on the quantum spin number
      // of the site in the UNIT CELL index `length'
      
      if (numberFlips <= maxNumberFlips) {
	
	counter++;
	
	conf[counter][1]=nn[1];
	conf[counter][2]=nn[2];
	conf[counter][3]=nn[3];
	conf[counter][4]=nn[4];
	
	if (counter != length) {
	  printf("TERM 24: Error counter != length...\nExiting!!\n\n");
	  exit(1);
	}
	
	findDescriptorOfCluster(counter,conf,descriptor);
	descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
	
	temp_multiplier=2*0.5*UnitCell.s2[nn[4]];
	
	// coeff from \tilde s^+ \tilde s^-
	
//	if (descriptorIndex > 0)
//	  sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
//	    hh[term].coefficient[index][neighbour];	
	
	if (Control.file) {
	  if (descriptorIndex > 0){
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=24;
	    indices[2]=descriptorIndex;
	    indices[3]=0;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=temp_multiplier*hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	  }
	}
	
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}  

double term25(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  /* TERM 25: 

     - F(k) . F(k) . S(+)(k) . S(+)(m).
     
  */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int tempString[5];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // F(k)
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k1][m]; // F(k)

	for(m=1;m<5;m++) 
	  tempString[m]=configuration[k2][m]; // s(+)(m)
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      

	for(j=1;j<=hh[term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* k ---> m */
	  
	  nn[1]=conf1[1][1]+hh[term].k_to_m[index][j][0];
	  nn[2]=conf1[1][2]+hh[term].k_to_m[index][j][1];
	  nn[3]=conf1[1][3]+hh[term].k_to_m[index][j][2];
	  nn[4]=hh[term].k_to_m[index][j][3];
	  
	  if(tempString[1] == nn[1] &&
	     tempString[2] == nn[2] &&
	     tempString[3] == nn[3] &&
	     tempString[4] == nn[4])
	    are_they_neighbours=j;
	  
	  if (are_they_neighbours) break;
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }	    
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=1;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    findDescriptorOfCluster(conf1_len,
					  conf1,
					  descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex1 <= 0) continue;
	    findDescriptorOfCluster(conf2_len,
					  conf2,
					  descriptor);
	    
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
	    
                if (descriptorIndex2 <= 0) continue;
	    // multiplier=1.0;
	    
	    // for(i=1;i<conf1_len;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf1_len;
	    
	    // configuration 1
	    // (length-1)! -- permutation factor because we sum over k.
	    // length      -- factor from F(k)
	    
	    // for(i=1;i<conf2_len;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf2_len;
	    
	    // configuration 2
	    // (length-1)! -- permutation factor because we sum over k
	    // length      -- factor from F(k)
	    
	    multiplier=-1.0;
	    
	    // coeff from \tilde s^+ \tilde s^-
	    
//	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//	      sum+=multiplier
//		*CCMcoeffs[descriptorIndex1]
//		*CCMcoeffs[descriptorIndex2]*
//		hh[term].coefficient[index][are_they_neighbours];
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=25;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term26(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 26: 
     
     S(+)(k) . S(+)(m) 

     */
  
  int tempConfig1[5];
  int tempConfig2[5];
  int nn[5];
  int j,k,m;
  int are_they_neighbours;
  int index;
  double sum;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  sum=0.0000;

  if(length == 2) {

    for(k=1;k<=2;k++) {
      
      if (k==1) {
	
	for(m=1;m<5;m++) 
	  tempConfig1[m]=string[1][m]; 
	
	for(m=1;m<5;m++) 
	  tempConfig2[m]=string[2][m]; 
	
      } else {

	for(m=1;m<5;m++) 
	  tempConfig1[m]=string[2][m]; 
	
	for(m=1;m<5;m++) 
	  tempConfig2[m]=string[1][m]; 
      }	

      // loop over nearest neighbours
    
      are_they_neighbours=0;
      index=tempConfig1[4];      
	
      for(j=1;j<=hh[term].k_to_m_number[index];j++) {
	
	// nearest neighbour is at x,y,z, temp put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=tempConfig1[1]+hh[term].k_to_m[index][j][0];
	nn[2]=tempConfig1[2]+hh[term].k_to_m[index][j][1];
	nn[3]=tempConfig1[3]+hh[term].k_to_m[index][j][2];
	nn[4]=hh[term].k_to_m[index][j][3];
	
	if(tempConfig2[1] == nn[1] &&
	   tempConfig2[2] == nn[2] &&
	   tempConfig2[3] == nn[3] &&
	   tempConfig2[4] == nn[4])
	  are_they_neighbours=j;     
	if (are_they_neighbours) break;
      }
      
      if (are_they_neighbours)
	sum+=1.0*hh[term].coefficient[index][are_they_neighbours];
      
      if (are_they_neighbours && Control.file) {
	int indices[7]={0};
	double dummy;
	indices[0]=config_index;
	indices[1]=26;
	indices[2]=0;
	indices[3]=0;
	indices[4]=0;
	indices[5]=0;
	dummy=1.0*hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
      }
      
    }
  }
  return sum;
}

double term27(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 27: 
     
     4 * H(k)(m) . S(z)(k) . S(z)(m)

     one of the most time-consuming terms because you've got to sweep
     through the index `k'

     */

  int i, j, l;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
//  double multiplier;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int minX, minY, minZ;
  int maxX, maxY, maxZ;

  int x,y,z,cell;
  int index;

  int numberFlips, maxNumberFlips;

  int (*configuration)[5];

  // term27 generates clusters of length+1+1, so we can skip some work
  // for length+1+1 > Params.n we always have an invalid index
  if (length+1+1 > Params.n) return 0;

  configuration=(int (*)[5])malloc((MAX_LSUBM+4)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  int maxPointX;
  int maxPointY;
  int maxPointZ;

  maxPointX=0;
  maxPointY=0;
  maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1])) maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2])) maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3])) maxPointZ=abs(string[j][3]);
  }

  if (Params.dim == 1) {
    minX=radius;           minY=0;                minZ=0;
    maxX=radius+maxPointX; maxY=0;                maxZ=0;
  } else if (Params.dim == 2) {
    minX=radius;           minY=radius;           minZ=0;
    maxX=radius+maxPointX; maxY=radius+maxPointY; maxZ=0;
  } else if (Params.dim == 3) {
    minX=radius;           minY=radius;           minZ=radius;
    maxX=radius+maxPointX; maxY=radius+maxPointY; maxZ=radius+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d)"
     " not recognised...\nExiting||\n\n",Params.dim);
    exit(1);
  }

//  multiplier=1.0;

  //  for(i=1;i<=length;i++)
  //  multiplier=multiplier*i;
  // multiplier=multiplier*(length+2)*(length+1);

  // length!               -- permutation factor
  // (length+2)*(length+1) -- factor from H(k,m)

  for(x=-minX;x<=maxX;x++) { // ToDo: better (faster) x = maxX-m ... m ???
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {

          // Joerg ToDo: lot of x,y,z pairs lying outside the bounding triangle
          //      and scipt if descriptor not exist, can be done earlier?
          //      same for term30
	  
	  configuration[length+1][1]=x;
	  configuration[length+1][2]=y;
	  configuration[length+1][3]=z;
	  configuration[length+1][4]=cell;
	  
	  index=cell;
	  numberFlips=1;
	  
	  for(l=1;l<=length;l++)
	    if (configuration[length+1][1]==configuration[l][1] &&
		configuration[length+1][2]==configuration[l][2] &&
		configuration[length+1][3]==configuration[l][3] &&
		configuration[length+1][4]==configuration[l][4])
	      numberFlips++;
	  
	  maxNumberFlips=UnitCell.s2[configuration[length+1][4]];
	  
	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	  
	  if (numberFlips <= maxNumberFlips) {
	    
	    for(j=1;j<=hh[term].k_to_m_number[index];j++) {
	      

	      // nearest neighbour is at x,y,z, temp put information
	      // into configuration and exit.
	      
	      /* k --> m */
	      
	      configuration[length+1+1][1]=configuration[length+1][1]
		+hh[term].k_to_m[index][j][0];
	      configuration[length+1+1][2]=configuration[length+1][2]
		+hh[term].k_to_m[index][j][1];
	      configuration[length+1+1][3]=configuration[length+1][3]
		+hh[term].k_to_m[index][j][2];
	      configuration[length+1+1][4]=
		hh[term].k_to_m[index][j][3];
	      
	      numberFlips=1;
	      
	      for(l=1;l<=length+1;l++)
		if (configuration[length+1+1][1]==configuration[l][1] &&
		    configuration[length+1+1][2]==configuration[l][2] &&
		    configuration[length+1+1][3]==configuration[l][3] &&
		    configuration[length+1+1][4]==configuration[l][4])
		  numberFlips++;
	      
	      maxNumberFlips=UnitCell.s2[configuration[length+1+1][4]];
	      
	      // the number of spin flips that you can have at a given
	      // lattice site is depedent on the quantum spin number
	      // of the site in the UNIT CELL index `length'
	      
	      if (numberFlips <= maxNumberFlips) {
		
		temp_multiplier=4*
		  0.5*UnitCell.s2[configuration[length+1][4]]*
		  0.5*UnitCell.s2[configuration[length+1+1][4]];
		
		findDescriptorOfCluster((length+1+1), configuration,
					      descriptor);
                // may be invalid if outside bounding box
		descriptorIndex=findDescriptorIndex(descriptor,
						    NumberConfigurations);

                // for length+1+1 > Params.n we always have an invalid index

//                // and must be exist as config
//		if (descriptorIndex > 0)
//		  sum+=temp_multiplier*CCMcoeffs[descriptorIndex]
//		    *hh[term].coefficient[index][j];
		
		if (Control.file) {
		  if (descriptorIndex > 0){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=27;
		    indices[2]=descriptorIndex;
		    indices[3]=0;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=temp_multiplier*
		      hh[term].coefficient[index][j];
            collect_term(indices,dummy,h_id);
		  }
		}
		
	      }
	    }
	  }
	}
      }
    }
  }
  return sum;
}

double term28(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  /* TERM 28: 

     2 * H(k)(m) . H(k)(m) . S(+)(k) . S(+)(m).
     
  */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // (k)

	for(m=1;m<5;m++) 
	  conf1[2][m]=configuration[k2][m]; // (k)
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k1][m]; // (m)
	
	for(m=1;m<5;m++) 
	  conf2[2][m]=configuration[k2][m]; // (m)
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      
	
	for(j=1;j<=hh[term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  nn[1]=conf1[1][1]+hh[term].k_to_m[index][j][0];
	  nn[2]=conf1[1][2]+hh[term].k_to_m[index][j][1];
	  nn[3]=conf1[1][3]+hh[term].k_to_m[index][j][2];
	  nn[4]=hh[term].k_to_m[index][j][3];
	  
	  if(conf1[2][1] == nn[1] &&
	     conf1[2][2] == nn[2] &&
	     conf1[2][3] == nn[3] &&
	     conf1[2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=2;
	    conf2_len=2;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    findDescriptorOfCluster(conf1_len,
					  conf1,
					  descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex1 <= 0) continue;
	    findDescriptorOfCluster(conf2_len,
					  conf2,
					  descriptor);
	    
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex2 <= 0) continue;
	    multiplier=1.0;
	    
	    // for(i=1;i<=(conf1_len-2);i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf1_len*
	    // (conf1_len-1);
	    
	    // configuration 1
	    // (length-2)!            -- permutation factor because 
	    //                           we sum over k and m.
	    // length*(length-1)      -- factor from H(k,m).
	    
	    // for(i=1;i<=(conf2_len-2);i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf2_len*
	    //  (conf2_len-1);
	      
	    // configuration 2
	    // (length-2)!            -- permutation factor because 
	    //                           we sum over k and m.
	    // length*(length-1)      -- factor from H(k,m).
	    
	    multiplier=2.0;
	    
	    // from \tilde s^- \tilde s^-
	    
//	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//	      sum+=multiplier
//		*CCMcoeffs[descriptorIndex1]
//		*CCMcoeffs[descriptorIndex2]*
//		hh[term].coefficient[index][are_they_neighbours];
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=28;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*
		  hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term29(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 29:
     
     4 * H(k)(m) . F(m) . S(+)(m) . S(z)(k)
     
  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;

  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  }

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++) {
      conf1[1][m]=configuration[k][m]; // F(m)
      conf2[1][m]=configuration[k][m]; // H(m,k) <-- m here.
    }
      
    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(m)
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // H(k,m)
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf2[1][4];
 
      for(neighbour=1;neighbour<=hh[term].m_to_k_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* m ---> k */
	
	nn[1]=conf2[1][1]+hh[term].m_to_k[index][neighbour][0];
	nn[2]=conf2[1][2]+hh[term].m_to_k[index][neighbour][1];
	nn[3]=conf2[1][3]+hh[term].m_to_k[index][neighbour][2];
	nn[4]=hh[term].m_to_k[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_len;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_len+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  findDescriptorOfCluster(conf1_len,
					conf1,
					descriptor);
	  descriptorIndex1=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	  
                if (descriptorIndex1 <= 0) continue;
	  findDescriptorOfCluster(temp_number_conf2_sites,
					conf2,
					descriptor);
	  
	  descriptorIndex2=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	  
                if (descriptorIndex2 <= 0) continue;
	  multiplier=1.0;

	  // for(i=1;i<conf1_len;i++)
	  //  multiplier=multiplier*i; 
	  // multiplier=multiplier*conf1_len;
	  
	  // F(k)
	  // (length-1)!       -- because we sum over k
	  // length            -- internal factor in F(k)
	  
	  //  for(i=1;i<=(temp_number_conf2_sites-2);i++)
	  //  multiplier=multiplier*i;
	  // multiplier=multiplier*temp_number_conf2_sites*
	  //  (temp_number_conf2_sites-1); 
	  
	  // H(k,m)
	  // (length-2)!       -- permutation factor from H(k,m)
	  // length*(length-1) -- internal factor in H(k,m) 
	  
	  multiplier=(-4)*0.5*UnitCell.s2[nn[4]]; // as nn[] <-- k
	  
	  // coefficient from \tilde s^- \tilde s^-
	  
//	  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//	    sum+=multiplier
//	      *CCMcoeffs[descriptorIndex1]
//	      *CCMcoeffs[descriptorIndex2]*
//	      hh[term].coefficient[index][neighbour];
	  
	  if (Control.file) {
	    if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=29;
	      indices[2]=descriptorIndex1;
	      indices[3]=descriptorIndex2;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*
		hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	    }
	  }
	  
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

/* this is the slowest term, 75% of CPU (Okt2004, Joerg) */
double term30(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 30: 
     
     4 * F(k) . F(m) . S(z)(k) . S(z)(m)

     one of the most time-consuming terms because you've got to sweep
     through the index `k'

     */
  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);


  int i, j, k, l, m;
  int descriptor[MAX_SPIN_LIST];
  double multiplier;

  int maxX, maxY, maxZ;
  int minX, minY, minZ;

  int x,y,z,cell;
  int maxPointX, maxPointY, maxPointZ;
  int index;

  int numberFlips, maxNumberFlips;

  int (*configuration)[5];

  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int counter;
  int spin[MAX_LSUBM+2];
  int neighbour;
  int temp_x;
  int temp_y;
  int temp_z;
  int temp_cell;
  int temp_number_conf1_sites;
  int temp_number_conf2_sites;
  int descriptorIndex1;
  int descriptorIndex2;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  maxPointX=0;
  maxPointY=0;
  maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1])) maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2])) maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3])) maxPointZ=abs(string[j][3]);
  }

  if (Params.dim == 1) { /* Joerg: change to minR[3] */
    minX=radius;           minY=0;                minZ=0;
    maxX=radius+maxPointX; maxY=0;                maxZ=0;
  } else if (Params.dim == 2) {
    minX=radius;           minY=radius;           minZ=0;
    maxX=radius+maxPointX; maxY=radius+maxPointY; maxZ=0;
  } else if (Params.dim == 3) {
    minX=radius;           minY=radius;           minZ=radius;
    maxX=radius+maxPointX; maxY=radius+maxPointY; maxZ=radius+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d) not recognised...\n"
           "Exiting||\n\n",Params.dim);
    exit(1);
  }

  for(x=-minX;x<=maxX;x++) {
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {
	  
          // ToDo: skip x+y+z outside triangle range [0..m]?

	  index=cell;

	  counter=length;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);   /* convert l to binary string spin[] */

            // spin[j]: 0 = F(m)  1 = F(k)
            // copy sites for 0-bit to conf2 and sites for 1-bit to conf1
	    conf1_len=0;
	    conf2_len=0;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		conf1_len++;
		for (m=1;m<5;m++) 
		  conf1[conf1_len][m]=configuration[j][m];
	      } else {
		conf2_len++;
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[j][m];
	      }
	    }
	    
	    numberFlips=1;
	    
	    for(m=1;m<=conf1_len;m++)
	      if (x==conf1[m][1] &&
		  y==conf1[m][2] &&
		  z==conf1[m][3] &&
		  cell==conf1[m][4])
		numberFlips++;
	    
	    maxNumberFlips = UnitCell.s2[cell];
	    
	    // the number of spin flips that you can have at a given
	    // lattice site is depedent on the quantum spin number
	    // of the site in the UNIT CELL index `length'
	    
	    if (numberFlips > maxNumberFlips) continue; 

            // Took this outside inner loop (Joerg Sep2005)
            temp_number_conf1_sites = conf1_len+1;
            conf1[temp_number_conf1_sites][1]=x;
            conf1[temp_number_conf1_sites][2]=y;
            conf1[temp_number_conf1_sites][3]=z;
            conf1[temp_number_conf1_sites][4]=cell;
            
            findDescriptorOfCluster(temp_number_conf1_sites,
                                          conf1, descriptor);
            descriptorIndex1=findDescriptorIndex(descriptor,
                                                 NumberConfigurations);
            if (descriptorIndex1 <= 0) continue;
	      
            for(neighbour=1;
                neighbour<=hh[term].k_to_m_number[cell];
                neighbour++) {
              
              // nearest neighbour is at x,y,z, cell put information
              // into configuration and exit.
              
              /* k ---> m */
              
              temp_x=x+hh[term].k_to_m[cell][neighbour][0];
              temp_y=y+hh[term].k_to_m[cell][neighbour][1];
              temp_z=z+hh[term].k_to_m[cell][neighbour][2];
              temp_cell=
                       hh[term].k_to_m[cell][neighbour][3];
              
              numberFlips=1;
              
              for(m=1;m<=conf2_len;m++)
                if (temp_x==conf2[m][1] &&
                    temp_y==conf2[m][2] &&
                    temp_z==conf2[m][3] &&
                    temp_cell==conf2[m][4])
                  numberFlips++;
              
              maxNumberFlips=UnitCell.s2[temp_cell];
              
              // the number of spin flips that you can have at a given
              // lattice site is depedent on the quantum spin number
              // of the site in the UNIT CELL index `length'
              
              if (numberFlips <= maxNumberFlips) {
                
                

                multiplier=1.0;
                // for(i=1;i<temp_number_conf1_sites;i++)
                //  multiplier=multiplier*i;
                // multiplier=multiplier*temp_number_conf1_sites;
                
                // (length-1)!           -- permutation factor sum over k
                // length                -- factor from F(k)
                
                temp_number_conf2_sites = conf2_len+1;
                conf2[temp_number_conf2_sites][1]=temp_x;
                conf2[temp_number_conf2_sites][2]=temp_y;
                conf2[temp_number_conf2_sites][3]=temp_z;
                conf2[temp_number_conf2_sites][4]=temp_cell;

                // for(i=1;i<temp_number_conf2_sites;i++)
                //  multiplier=multiplier*i;
                // multiplier=multiplier*temp_number_conf2_sites;
                
                // (length-1)!           -- permutation factor sum over m
                // length                -- factor from F(m)
                
                multiplier=4*0.5*UnitCell.s2[cell]*
                             0.5*UnitCell.s2[temp_cell];
                
                // coeff from \tilde s^- \tilde s^- 
                
                findDescriptorOfCluster(temp_number_conf2_sites,
                                              conf2, descriptor);
                descriptorIndex2=findDescriptorIndex(descriptor,
                                                     NumberConfigurations);
                if (descriptorIndex2 <= 0) continue;

//                if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//                  sum+=multiplier*CCMcoeffs[descriptorIndex1]
//                                 *CCMcoeffs[descriptorIndex2]*
 //                   hh[term].coefficient[index][neighbour];
                
                if (Control.file) {
                  if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
                    int indices[7]={0};
                    double dummy;
                    indices[0]=config_index;
                    indices[1]=30;
                    indices[2]=descriptorIndex1;
                    indices[3]=descriptorIndex2;
                    indices[4]=0;
                    indices[5]=0;
                    dummy=multiplier*
                      hh[term].coefficient[index][neighbour];
                    collect_term(indices,dummy,h_id);
                  }
                }
                
              }
            }
	  }
	}
      }
    }
  }
  return sum;
}

double term31(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 31: 
     
     2 * F(m) . F(m) . F(k) . S(+)(m) . S(z)(k)   (like term33)
   
    */
  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);


  int spin[MAX_LSUBM+1];
  int mm,nnn,temp1,temp2,temp3;
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int i, j, k, l, m;
  int nn[5];
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
//  int temp_number_conf3_sites;
  int numberFlips;
  int maxNumberFlips;
  int (*configuration)[5];
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k=1;k<=length;k++) {
    
    for(m=1;m<5;m++) conf1[1][m]=configuration[k][m]; // F(m)
    for(m=1;m<5;m++) conf2[1][m]=configuration[k][m]; // F(m)
    
    // loop over nearest neighbours
    
    counter=1;
    for(l=1;l<=length;l++) {
      if (l != k) {
	tempArray[counter]=l;
	counter++;
      }
    }
    
    counter--;
    // see term33
    
    if (counter > 30) {
      printf("TERM 31: counter > 30...\nExiting!!\n\n");
      exit(1);
    }
    temp = counter; // split into summands not greater than 10 each
    temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
    temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
    temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
    
    for(l=0;l<trimary_max[temp1];l++) {
      for(mm=0;mm<trimary_max[temp2];mm++) {
	for(nnn=0;nnn<trimary_max[temp3];nnn++) {
	  
	  trimary(l,mm,nnn,counter,spin);
	  conf1_len=1;
	  conf2_len=1;
	  conf3_len=0;
	  
	  for (j=1;j<=counter;j++) { 
	    temp=tempArray[j];
	    if (spin[j]==0) {         // F(m)
	      conf1_len++;
	      for (m=1;m<5;m++) 
		conf1[conf1_len][m]=configuration[temp][m];
	    } else if (spin[j]==1) {  // F(m)
	      conf2_len++;
	      for (m=1;m<5;m++) 
		conf2[conf2_len][m]=configuration[temp][m];
	    } else {                  // F(k)
	      conf3_len++;
	      for (m=1;m<5;m++) 
		conf3[conf3_len][m]=configuration[temp][m];
	    }
	  }	      
	  

          if (!subsetClusters(conf1,conf1_len))
              continue;
          if (!subsetClusters(conf2,conf2_len))
              continue;
          // moved that outside inner loop, Joerg Sep05, speedup
          findDescriptorOfCluster(conf1_len,
                                        conf1, descriptor);
          descriptorIndex1=findDescriptorIndex(descriptor,
                                               NumberConfigurations);
          if (descriptorIndex1 <= 0) continue;

          findDescriptorOfCluster(conf2_len,
                                        conf2, descriptor);
          descriptorIndex2=findDescriptorIndex(descriptor,
                                               NumberConfigurations);
          if (descriptorIndex2 <= 0) continue;
	  
	  index=conf1[1][4];      
	  for(j=1;j<=hh[term].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+hh[term].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+hh[term].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+hh[term].m_to_k[index][j][2];
	    nn[4]=hh[term].m_to_k[index][j][3];
	    
	    numberFlips=1;
	    
	    for(m=1;m<=conf3_len;m++)
	      if (nn[1]==conf3[m][1] &&
		  nn[2]==conf3[m][2] &&
		  nn[3]==conf3[m][3] &&
		  nn[4]==conf3[m][4])
		numberFlips++;
	    
	    maxNumberFlips=UnitCell.s2[nn[4]];
	    
	    // the number of spin flips that you can have at a given
	    // lattice site is depedent on the quantum spin number
	    // of the site in the UNIT CELL index `length'
	    
	    if (numberFlips <= maxNumberFlips) {
	      
	      // temp_number_conf3_sites = conf3_len+1;
	      conf3[conf3_len+1][1]=nn[1];
	      conf3[conf3_len+1][2]=nn[2];
	      conf3[conf3_len+1][3]=nn[3];
	      conf3[conf3_len+1][4]=nn[4];
	      
	      findDescriptorOfCluster(conf3_len+1, conf3, descriptor);
	      descriptorIndex3=findDescriptorIndex(descriptor,
						   NumberConfigurations);
	      if (descriptorIndex3 <= 0) continue;
	      
              multiplier=1.0;
	      
	      // for(i=1;i<conf1_len;i++)
	      // multiplier=multiplier*i;
	      // multiplier=multiplier*conf1_len;
	      
	      // F(m)
	      // in each case:
	      // (length-1)!      -- we sum over k.
	      // length           -- internal factor in F().
	      
	      // for(i=1;i<conf2_len;i++)
	      // multiplier=multiplier*i;
	      // multiplier=multiplier*conf2_len;
	      
	      // F(m)
	      // in each case:
	      // (length-1)!      -- we sum over k.
	      // length           -- internal factor in F().
	      
		
	      //	      for(i=1;i<conf3_len+1;i++)
	      //	multiplier=multiplier*i;
	      // multiplier=multiplier*(conf3_len+1);
	      
	      // F(k)
	      // (length-1)!      -- we sum over nearest-neihbours (m).
	      // length           -- internal factor in F().
	      
	      multiplier=(-2)*0.5*UnitCell.s2[nn[4]]; // as nn[] <-- k
	      
	      // from \tilde s^- \tilde s^-
	      
//	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)
//		  && (descriptorIndex3 > 0))
//		  sum+=multiplier
//		    *CCMcoeffs[descriptorIndex1]
//		    *CCMcoeffs[descriptorIndex2]
//		    *CCMcoeffs[descriptorIndex3]*
//		    hh[term].coefficient[index][j];
	      
	      if (Control.file) {
		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
		    (descriptorIndex3 > 0)){
		  int indices[7]={0};
		  double dummy;
		  indices[0]=config_index;
		  indices[1]=31;
		  indices[2]=descriptorIndex1;
		  indices[3]=descriptorIndex2;
		  indices[4]=descriptorIndex3;
		  indices[5]=0;
		  dummy=multiplier*
		    hh[term].coefficient[index][j];
                  collect_term(indices,dummy,h_id);
		}
	      }
		
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}

double term32(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 32: 
     
     4 * F(k) . F(m) . H(k)(m) . S(+)(k) . S(+)(m)
   
    */
  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);


  int spin[MAX_LSUBM+1];
  int mm,nnn,temp1,temp2,temp3;
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; // F(k)
	for(m=1;m<5;m++) conf2[1][m]=configuration[k2][m]; // F(m)
	for(m=1;m<5;m++) conf3[1][m]=configuration[k1][m]; // H(k,m) <--- k
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      

	for(j=1;j<=hh[term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* k ---> m */
	  
	  nn[1]=conf1[1][1]+hh[term].k_to_m[index][j][0];
	  nn[2]=conf1[1][2]+hh[term].k_to_m[index][j][1];
	  nn[3]=conf1[1][3]+hh[term].k_to_m[index][j][2];
	  nn[4]=hh[term].k_to_m[index][j][3];
	  
	  if(conf2[1][1] == nn[1] &&
	     conf2[1][2] == nn[2] &&
	     conf2[1][3] == nn[3] &&
	     conf2[1][4] == nn[4])
	    are_they_neighbours=j;

	  if (are_they_neighbours) break;
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  for(m=1;m<5;m++) conf3[2][m]=configuration[k2][m]; // H(k,m) <--- m
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
		counter++;
	    }
	  }
	  
	  counter--;
	  
	  if (counter > 30) {
	    printf("TERM 32: counter > 30...\nExiting!!\n\n");
	    exit(1);
	  }
          temp = counter; // split into summands not greater than 10 each
          temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
          temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
          temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	  
	  for(l=0;l<trimary_max[temp1];l++) {
	    for(mm=0;mm<trimary_max[temp2];mm++) {
	      for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		
		trimary(l,mm,nnn,counter,spin);
		conf1_len=1;
		conf2_len=1;
		conf3_len=2;
		  
		for (j=1;j<=counter;j++) { 
		  temp=tempArray[j];
		  if (spin[j]==0) {          // F(k)
		    conf1_len++;
		    for (m=1;m<5;m++) 
		      conf1[conf1_len][m]=configuration[temp][m];
		  } else if (spin[j]==1) {   // F(m)
		    conf2_len++;
		    for (m=1;m<5;m++) 
		      conf2[conf2_len][m]=configuration[temp][m];
		  } else {                    // F(m)
		    conf3_len++;
		    for (m=1;m<5;m++) 
		      conf3[conf3_len][m]=configuration[temp][m];
		  }
		}	      
		
          if (!subsetClusters(conf1,conf1_len))
              continue;
          if (!subsetClusters(conf2,conf2_len))
              continue;
		findDescriptorOfCluster(conf1_len,
					      conf1, descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
                if (descriptorIndex1 <= 0) continue;

		findDescriptorOfCluster(conf2_len,
					      conf2, descriptor);
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
                if (descriptorIndex2 <= 0) continue;

		findDescriptorOfCluster(conf3_len,
					      conf3, descriptor);
		descriptorIndex3=findDescriptorIndex(descriptor,
						     NumberConfigurations);
                if (descriptorIndex3 <= 0) continue;

		multiplier=1.0;
		
		//		for(i=1;i<conf1_len;i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf1_len;
		
		// for(i=1;i<conf2_len;i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf2_len;
		
		// in each case above F():
		// (length-1)!      -- we sum over k1 and k2.
		// length           -- internal factor in F().
		
		// for(i=1;i<=(conf3_len-2);i++)
		//    multiplier=multiplier*i;
		// multiplier=multiplier*conf3_len*
		//  (conf3_len-1);
		
		// H(k,m)
		// (length-2)!        -- we sum over k1 and k2.
		// length*(length-1)  -- internal factor in H().
		
		multiplier=4.0;
		
		// from \tilde s^- \tilde s^-
		
//		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)
//		    && (descriptorIndex3 > 0))
//		  sum+=multiplier
//		    *CCMcoeffs[descriptorIndex1]
//		    *CCMcoeffs[descriptorIndex2]
//		    *CCMcoeffs[descriptorIndex3]*
//		    hh[term].coefficient[index][are_they_neighbours];
		  
		if (Control.file) {
		  if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)&&
		      (descriptorIndex3 > 0)){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=32;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=descriptorIndex3;
		    indices[5]=0;
		    dummy=multiplier*
		      hh[term].coefficient[index][are_they_neighbours];
                    collect_term(indices,dummy,h_id);
		  }
		}
		  
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}

double term33(int config_index, int conf0_len, int (*conf0)[5],
              struct hamiltonian_defn hh[], int term)
{
  
  /* TERM 33: 
     
     2 * F(k) . F(k) . F(m) . S(+)(k) . S(z)(m)  (like term31)

   split cfg using !k into 3cfgs, add k to cfg 1+2, add H?? to cfg3
   see SrCuBO-4-4 cfg6 to 6,3,1  3,6,1  4,5,1  5,4,1  4,3,2  3,4,2 
       len          3     3 1 1  1 3 1  2 2 1  2 2 1  2 1 2  1 2 2
       len(-1,-1,-0)      2 0 1  0 2 1  1 1 1  1 1 1  1 0 2  0 1 2  
    6= (0,0,0,4) (0,0,0,3) (0,0,0,2) \|
    5= (0,0,0,3) (0,0,0,2) |
    4= (0,0,0,4) (0,0,0,2) \
    3= (0,0,0,1)           .
    2= (0,0,0,4) (0,0,0,3) -
    1= (0,0,0,3)           '    # == (0,0,0,1)? missing sym?
   see SrCuBO-2-2 cfg3 to 3,3,1 l=1 split nothing, but cfg1=k,cfg2=k,cfg3=m
       len         1      1 1 1
       len(-1,-1,-0)      0 0 1
    3= (0,0,3)
    1= (0,0,1)
    sum to 0 for alpha=0 ?!!!
     
  */
  
  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  // int spin[MAX_LSUBM+1];
  static int split1[MAX_LSUBM+1];
  static int split2[MAX_LSUBM+1];
  int counter;
  int len2;
  // int temp;
  int (*conf1)[5], (*conf2)[5], (*conf3)[5], (*rest1)[5], (*rest2)[5];
  int  j, k, l, m;
  int nn[5];
//  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
//  int tempArray[MAX_LSUBM+1];
//  int temp_number_conf3_sites;
  int numberFlips;
  int maxNumberFlips;
  double sum;
  sum=0.0000;

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  rest1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  rest2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k=1;k<=conf0_len;k++) {
    
    for(m=1;m<5;m++) conf1[1][m]=conf0[k][m]; // F(k)
    for(m=1;m<5;m++) conf2[1][m]=conf0[k][m]; // F(k)
    
    // rest1 is conf0 without site k
    
    for(counter=0,l=1;l<=conf0_len;l++)
      if (l != k) {
        counter++;
        for(m=1;m<5;m++) rest1[counter][m]=conf0[l][m];
      } // counter is conf0_len-1
/*
   see SrCuBO-2-2 cfg3 to 3,3,1
       len         1      1 1 1
       len(-1,-1,-0)      0 0 1
    3= (0,0,3) len=1
    1= (0,0,1)
*/   
    /* decompose the rest spins to 3 configs (stepwise 2 times to 2 configs) */
    { int i; for (i=1;i<=conf0_len-1;i++) split1[i]=0; }
    int i1=0; // split counter overflow
    for(;;) {
      conf1_len=1; // first site is k before split0 is appended
      len2=0;      // rest is empty   before split1 is appended
      if (i1) break; 
      /* split configuration and count up (new: no index list) */
      i1= split_config(rest1, conf0_len-1, split1,
                       conf1, &conf1_len,
                       rest2, &len2);

      /* speedup Apr06 JS */
      if (!subsetClusters(conf1,conf1_len)) continue;
      // moved outside inner loop, Joerg Sep2005, speedup
      findDescriptorOfCluster(conf1_len, conf1, descriptor);
      descriptorIndex1=findDescriptorIndex(descriptor, NumberConfigurations);
      if (descriptorIndex1 <= 0) continue;
  if (Control.output > 8){
   printf("Term33 idx= %d len=%d k=%d\n",config_index,conf0_len,k);
   for( int i=1; i<=conf0_len; i++)
     printf(" %d x y uc %d %d %d\n",i,conf0[i][1], conf0[i][2], conf0[i][4]);
 }
  if (Control.output > 8){
   printf(" conf1  len=%d\n", conf1_len);
   for( int i=1; i<=conf1_len; i++)
     printf(" %d x y uc %d %d %d\n",i,conf1[i][1], conf1[i][2], conf1[i][4]);
   printf(" rest2  len=%d\n", len2);
   for( int i=1; i<=len2; i++)
     printf(" %d x y uc %d %d %d\n",i,rest2[i][1], rest2[i][2], rest2[i][4]);
 }

      { int i; for (i=1;i<=len2;i++) split2[i]=0; }
      int i2=0; // split counter overflow
      for(;;) {
        conf2_len=1; // first site is k before split0 is appended
        conf3_len=0; // empty           before split1 is appended
        if (i2) break;
        /* split configuration and count up (new: no index list) */
        i2= split_config(rest2, len2, split2,
                         conf2, &conf2_len,
                         conf3, &conf3_len);
	  
        if (!subsetClusters(conf2,conf2_len)) continue;
        findDescriptorOfCluster(conf2_len, conf2, descriptor);
        descriptorIndex2=findDescriptorIndex(descriptor, NumberConfigurations);
        if (descriptorIndex2 <= 0) continue;

  if (Control.output > 8){
   printf(" conf2  len=%d\n", conf2_len);
   for( int i=1; i<=conf2_len; i++)
     printf(" %d x y uc %d %d %d\n",i,conf2[i][1], conf2[i][2], conf2[i][4]);
   printf(" conf3  len=%d\n", conf3_len);
   for( int i=1; i<=conf3_len; i++)
     printf(" %d x y uc %d %d %d\n",i,conf3[i][1], conf3[i][2], conf3[i][4]);
 }

//	are_they_neighbours=0;
	index=conf1[1][4];   /* uc of site 1 */
        /* ToDo: make a subroutine! */
	for(j=1;j<=hh[term].k_to_m_number[index];j++) {
	    
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  /* k ---> m */
	  
	  nn[1]=conf1[1][1]+hh[term].k_to_m[index][j][0];
	  nn[2]=conf1[1][2]+hh[term].k_to_m[index][j][1];
	  nn[3]=conf1[1][3]+hh[term].k_to_m[index][j][2];
	  nn[4]=            hh[term].k_to_m[index][j][3];

          // ToDo: num_flips_check(conf3(+nn),conf3_len(+1))	    
	  numberFlips=1;
	  for(m=1;m<=conf3_len;m++)   // len = {0.. }
	    if (nn[1]==conf3[m][1] &&
		nn[2]==conf3[m][2] &&
		nn[3]==conf3[m][3] &&
		nn[4]==conf3[m][4])
		numberFlips++; // range=0.. 0 for l=1 cfgs
	  
	  maxNumberFlips=UnitCell.s2[nn[4]];
	    
	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `conf0_len'
	  
	  if (numberFlips <= maxNumberFlips) { // append k --> m
	      
	      // temp_number_conf3_sites=conf3_len+1;
	      conf3[conf3_len+1][1]=nn[1];
	      conf3[conf3_len+1][2]=nn[2];
	      conf3[conf3_len+1][3]=nn[3];
	      conf3[conf3_len+1][4]=nn[4];
  if (Control.output > 8)
   printf(" conf3  len=%d OK1\n", conf3_len+1);
	      
	      findDescriptorOfCluster(conf3_len+1, conf3, descriptor);
	      descriptorIndex3=findDescriptorIndex(descriptor,
						   NumberConfigurations);
              if (descriptorIndex3 <= 0) continue;
  if (Control.output > 8)
   printf(" conf3  len=%d OK2\n", conf3_len+1);

	      multiplier=1.0;
	      
	      // multiplier*=conf1_len;
	      // F(k)
	      // multiplier*=conf2_len;
	      // F(k)
	      // multiplier*=conf3_len+1;
	      //F(m)
	      
	      multiplier=(-2)*0.5*UnitCell.s2[nn[4]]; // as nn[] <-- m
	      
	      // from \tilde s^- \tilde s^-
	      
//	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)
//		  && (descriptorIndex3 > 0))
//		sum+=multiplier
//		  *CCMcoeffs[descriptorIndex1]
//		  *CCMcoeffs[descriptorIndex2]
//		  *CCMcoeffs[descriptorIndex3]*
//		  hh[term].coefficient[index][j];
	      
	      if (Control.file) {
		if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)&&
		    (descriptorIndex3 > 0)){
		  int indices[7]={0};
		  double dummy;
		  indices[0]=config_index;
		  indices[1]=33;
		  indices[2]=descriptorIndex1;
		  indices[3]=descriptorIndex2;
		  indices[4]=descriptorIndex3;
		  indices[5]=0;
		  dummy=multiplier*
		    hh[term].coefficient[index][j];
                  collect_term(indices,dummy,h_id);
		}
	      }
	      
	    }
	  }
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(rest1);
  free(rest2);
  return sum;
}

/* slowest 31% for chain-6-6 Joerg Sep05  new algo */
/* length = 1..MAX_LSUBM  */
double term34(int config_index, int length, int (*config)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 34:
     
     F(k) . F(k) . F(m) . F(m) . S(+)(k) . S(+)(m)
   
    */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  static int split1[MAX_LSUBM+1]; /* decompose bit pattern 1=take, 0=rest */
  static int split2[MAX_LSUBM+1]; /* decompose bit pattern 1=take, 0=rest */
  static int split3[MAX_LSUBM+1]; /* decompose bit pattern 1=take, 0=rest */
  int len1, len2, len3;
  // int temp;
  int (*conf[4])[5]; // Joerg: replaces old (*conf1)[5]..**conf4
  int (*rest1)[5], (*rest2)[5], (*rest3)[5];
  int k, l, m, mm;
  int k1, k2, ks;
  // int nn[5];
  int are_they_neighbours;
  int index;
  static int descriptor[MAX_SPIN_LIST];
  int descriptorIndex[4]; // Joerg: replace old descriptor1..4
  // double multiplier;
  int conf_len[4]; // Joerg: replaces conf[1..4]_len
  double sum;
  sum=0.0000;

  /* config: {{unused,},x,y,z,uc} */

  conf[0]=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));   // default: MAX_LSUB=20
  conf[1]=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf[2]=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf[3]=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  rest1  =(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  rest2  =(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  rest3  =(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++)
   for(l=1;l<5;l++) {
      conf[0][k][l]=0;
      conf[1][k][l]=0;
      conf[2][k][l]=0;
      conf[3][k][l]=0;
  } 
 
  for(k1=1;k1<=length;k1++)
   for(k2=1;k2<=length;k2++)
    // if (k1 < k2) {} // old: if (k2 != k1)  Sep05, use 12 <-> 34 symmetry
    // Buggy for SrCuO4-2-2 + Alpha 2-34-1-1-1-1 1.0 missing
    //   should be ok, if also for m-to-k neighbourhood is checked !?
    if (k1 != k2) { // old: if (k2 != k1)  Sep05, use 12 <-> 34 symmetry
    // k1 and k2 must fullfill restrictions! k1 from A than k2 from B-sublat.!
	
	for(m=1;m<5;m++) {
	  conf[0][1][m]=config[k1][m]; // F(k) 12 and 34 are symmetric!
	  conf[1][1][m]=config[k1][m]; // F(k)
	  conf[2][1][m]=config[k2][m]; // F(m)
	  conf[3][1][m]=config[k2][m]; // F(m)
        }
	
	// loop over nearest neighbours
	
        index = conf[0][1][4];
        are_they_neighbours=
           test_k_to_m_neighbour(hh,term,conf[0][1],conf[2][1]);
	// conf[0][1] and conf[2][1] need to be nearest-neighbours 
	
        // remove sites k1 and k2
	if (are_they_neighbours) {
	  
	  len1=0;
	  for(l=1;l<=length;l++) {     // k3
	    if (l != k1 && l != k2) {  // different from k1 and k2
	      len1++;
	      for (mm=1;mm<5;mm++) rest1[len1][mm] = config[l][mm];
	    }
	  }
	  
          // now we decompose configuration into 4 configurations
          //  ToDo: better do successiv decomposition into 2 configurations?
          //        than decompose 2nd config again into 2 parts
          //        1. decompose tmp  binary to Config(1) + rest
          //        2. decompose rest binary to Config(2) + tmp
          //        3. decompose tmp  binary to Config(3) + Config(4)
          // quartery code decomposed to dual + trimary code
          // 000 001 002 003 010 011 012 => 000 001 010 011 100 101
          //                                000 00  0 0
          //                                001 01  0 1
          //                                002 02  0 2
          //                                010 10  1 0
// ToDo: as function? config0 auf config1..n aufteilen (anhaengen)
{
          { int i; for (i=1;i<=len1;i++) split1[i]=0; }
          int i1=0; // counter overflow
          for(;;) {
	    conf_len[0]=1; // first site is k1 or k2
            len2=0;
            if (i1) break;
            /* split configuration and count up (new: no index list) */
            i1= split_config(rest1,   len1, split1,
                             conf[0], &conf_len[0],
                             rest2,   &len2);
            ks=0; {
                  // this consumes most of CPU time
                  //  square_s1-8-4 configs=30 calls=604004/1508970
                  //    decompositions: cfgs^4=30^4=810e3
                  // new Mar06 testwise (speed*2)
                  if (!subsetClusters(conf[ks],conf_len[ks]))
                    continue;
  		  findDescriptorOfCluster(
                    conf_len[ks],conf[ks],descriptor);
		  descriptorIndex[ks]=
                    findDescriptorIndex(descriptor,NumberConfigurations);
                  // output here testwise???
                  if (descriptorIndex[ks] <= 0) continue; // mostly invalid
                  // ToDo: better get only valid confs, count valid/invalid?
                }
              
          /* decompose the rest spins to other 3 configs */
          { int i; for (i=1;i<=len2;i++) split2[i]=0; }
          int i2=0; // counter overflow
          for(;;) {
	    conf_len[1]=1; // first site is k1 or k2 before split appended
            len3=0;
            if (i2) break;
            /* split configuration and count up (new: no index list) */
            i2= split_config(rest2,   len2, split2,
                             conf[1], &conf_len[1],
                             rest3,   &len3);
            ks=1; {
                  // this consumes most of CPU time
                  //  square_s1-8-4 configs=30 calls=604004/1508970
                  //    decompositions: cfgs^4=30^4=810e3
                  // new Mar06 testwise (speed*2)
                  if (!subsetClusters(conf[ks],conf_len[ks]))
                    continue;
  		  findDescriptorOfCluster(
                    conf_len[ks],conf[ks],descriptor);
		  descriptorIndex[ks]=
                    findDescriptorIndex(descriptor,NumberConfigurations);
                  // output here testwise???
                  if (descriptorIndex[ks] <= 0) continue; // mostly invalid
                  // ToDo: better get only valid confs, count valid/invalid?
                }
              
          /* decompose the rest spins to other 2 configs */
          { int i; for (i=1;i<=len3;i++) split3[i]=0; }
          int i3=0; // counter overflow
          for(;;) {
	    conf_len[2]=1; // first site is k1 or k2
            conf_len[3]=1;
            if (i3) break;
            /* split configuration and count up (new: no index list) */
            i3=  split_config(rest3,   len3, split3,
                              conf[2], &conf_len[2],
                              conf[3], &conf_len[3]);
                // slowest part
                //  check if we have 4 valid configs conf[] here
                for (ks=2; ks<4; ks++) {
                  // this consumes most of CPU time
                  //  square_s1-8-4 configs=30 calls=604004/1508970
                  //    decompositions: cfgs^4=30^4=810e3
                  // new Mar06 testwise (speed*2)
                  if (!subsetClusters(conf[ks],conf_len[ks]))
                    break;
  		  findDescriptorOfCluster(
                    conf_len[ks],conf[ks],descriptor);
		  descriptorIndex[ks]=
                    findDescriptorIndex(descriptor,NumberConfigurations);
                  // output here testwise???
                  if (descriptorIndex[ks] <= 0) break; // mostly invalid
                  // ToDo: better get only valid confs, count valid/invalid?
                }

                // Joerg ToDo: skip     if descrIndex[2]>descrIndex[1]
                //             factor 2 if descrIndex[2]<descrIndex[1]
                if (ks<4) continue; // skip the rest

                // Joerg: now we have 4 valid indices to configs
                //        its probably faster to test NumConfigs^4 (<=length-1)
                //        if they build our config (length, one same site + miss side)
                // LC-4-4 no term
                // LC-6-6 only Config9: 16*(5,2,2,1)+24*(2,2,1,1)
                // oooooo -> o....o o..o o..o oo + o..o o..o oo oo
                //  sum_len(cfg1..4)=length(conf9)+2 ???
                //  first two configs and last two configs are symmetric!
                //        -> always factor 8 (divide to 12 and 34 and fact=2)
                //  we test n^2*4^(n-2) decompositions here (grows exponentially)
                //   

		// multiplier=1.0;
		// multiplier = * (conf_len[0]-1)!
                //              * (conf_len[1]-1)!
                //              * (conf_len[2]-1)!
                //              * (conf_len[3]-1)!
		// in each case:
		// (length-1)!      -- we sum over k1 and k2.
		// length           -- internal factor in F().
		
//                // ToDo: CCMcoeff=0.0 && sum not used? remove? Joerg sep05
//		sum+=CCMcoeffs[descriptorIndex[0]]
//		    *CCMcoeffs[descriptorIndex[1]]
//		    *CCMcoeffs[descriptorIndex[2]]
//		    *CCMcoeffs[descriptorIndex[3]]
//		//  *2
 //          *hh[term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		    static int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;  // original config
		    indices[1]=34;
		    indices[2]=descriptorIndex[0]; // modified config1..4
		    indices[3]=descriptorIndex[1];
		    indices[4]=descriptorIndex[2];
		    indices[5]=descriptorIndex[3];
		    dummy= // 2*  // factor 2 for 3412 sym, Joerg Sep05
		     hh[term].coefficient[index][are_they_neighbours];
                    collect_term(indices,dummy,h_id);
		}
		
	      }
	    }
	  }
          }
	}
  }
  free(conf[0]);
  free(conf[1]);
  free(conf[2]);
  free(conf[3]);
  free(rest1  );
  free(rest2  );
  free(rest3  );
  return sum;  // not used!
}



double term35(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  
  /* TERM 35:
     
     4 * H(k)(m) . F(k) . S(+)(k) . S(z)(m)
     
  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++) {
      conf1[1][m]=configuration[k][m]; // F(k)
      conf2[1][m]=configuration[k][m]; // H(m,k) <-- k here.
    }
      
    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(k)
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // H(k,m)
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf2[1][4];
      
      for(neighbour=1;neighbour<=hh[term].k_to_m_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=conf2[1][1]+hh[term].k_to_m[index][neighbour][0];
	nn[2]=conf2[1][2]+hh[term].k_to_m[index][neighbour][1];
	nn[3]=conf2[1][3]+hh[term].k_to_m[index][neighbour][2];
	nn[4]=            hh[term].k_to_m[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_len;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_len+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  findDescriptorOfCluster(conf1_len,
					conf1,
					descriptor);
	  descriptorIndex1=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	  
                if (descriptorIndex1 <= 0) continue;
	  findDescriptorOfCluster(temp_number_conf2_sites,
					conf2,
					descriptor);
	  
	  descriptorIndex2=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	    
                if (descriptorIndex2 <= 0) continue;
	  multiplier=1.0;
	  // for(i=1;i<conf1_len;i++)
	  //  multiplier=multiplier*i; 
	  // multiplier=multiplier*conf1_len;
	    
	  // F(k)
	  // (length-1)!       -- because we sum over k
	  // length            -- internal factor in F(k)
	  
	  // for(i=1;i<=(temp_number_conf2_sites-2);i++)
	  //  multiplier=multiplier*i;
	  // multiplier=multiplier*temp_number_conf2_sites*
	  //  (temp_number_conf2_sites-1); 
	  
	  // H(k,m)
	  // (length-2)!       -- permutation factor from H(m)
	  // length*(length-1) -- internal factor in H(k,m) 
	  
	  multiplier=(-4)*0.5*UnitCell.s2[nn[4]]; // as nn[] <-- m
	  
	  // coefficient from \tilde s^- \tilde s^-
	    
//	  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//	    sum+=multiplier
//	      *CCMcoeffs[descriptorIndex1]
//	      *CCMcoeffs[descriptorIndex2]*
//	      hh[term].coefficient[index][neighbour];
	    
	  if (Control.file) {
	    if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=35;
	      indices[2]=descriptorIndex1;
	      indices[3]=descriptorIndex2;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*
		hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	    }
	  }
	  
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

// now the ``extra'' terms for s > 1/2.

double term36(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 36:
     
     2 * M(k)(k)(m) . S(+)(k) . S(z)(m)
     
  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  int (*configuration)[5];
  int (*conf)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips1;
  int numberFlips2;
  int temp_length;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;

  //  multiplier=(length+1.0);

  //  for(i=1;i<length;i++)
  //  multiplier=multiplier*i; 

  // multiplier=multiplier*(length+2)*(length+1)*length; 

  // (length-1)!        -- permutation factor as we sum over k
  // (length+1)*length  -- internal factor from H(). Note that 
  //                       summation over n.n. gives m thus 
  //                       length ->length+1.

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];

    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++)
	  conf[j][m]=configuration[j][m]; // M(k,k,m)
      
      temp_length=length+1;
      
      // we repeat the `k' 
      
      for(m=1;m<5;m++)
	conf[temp_length][m]=configuration[k][m];

      index=configuration[k][4];
      
      for(neighbour=1;neighbour<=hh[term].k_to_m_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=configuration[k][1]+
	  hh[term].k_to_m[index][neighbour][0];
	nn[2]=configuration[k][2]+
	  hh[term].k_to_m[index][neighbour][1];
	nn[3]=configuration[k][3]+
	  hh[term].k_to_m[index][neighbour][2];
	nn[4]=hh[term].k_to_m[index][neighbour][3];
	
	numberFlips1=1;
	
	for(l=1;l<=temp_length;l++)
	  if (nn[1]==conf[l][1] &&
	      nn[2]==conf[l][2] &&
	      nn[3]==conf[l][3] &&
	      nn[4]==conf[l][4])
	    numberFlips1++;
	
	numberFlips2=0;
	
	for(l=1;l<=temp_length;l++)
	  if (configuration[k][1]==conf[l][1] &&
	      configuration[k][2]==conf[l][2] &&
	      configuration[k][3]==conf[l][3] &&
	      configuration[k][4]==conf[l][4])
	    numberFlips2++;
	
	maxNumberFlips1=UnitCell.s2[nn[4]];
	maxNumberFlips2=UnitCell.s2[index];

	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips1 <= maxNumberFlips1 && 
	    numberFlips2 <= maxNumberFlips2)  {
	  
	  temp_length=length+2;
	  
	  conf[temp_length][1]=nn[1]; // `m' is the nearest neighbour.
	  conf[temp_length][2]=nn[2];
	  conf[temp_length][3]=nn[3];
	  conf[temp_length][4]=nn[4];
	  
	  findDescriptorOfCluster(temp_length,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  temp_multiplier=-2*0.5*UnitCell.s2[nn[4]];
	  
	  // coefficient from \tilde s^- \tilde s^- expansion
	  
//	  if (descriptorIndex > 0)
//	    sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
//		hh[term].coefficient[index][neighbour];	
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {	
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=36;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	    }
	  }

	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}
  
double term37(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 37:
     
     2 * M(k)(m)(m) . S(+)(m) . S(z)(z)
     
     NOTE: DIRECTIONALITY TO BE ADDED FOR TRIANGLE!

  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  // double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips1;
  int numberFlips2;
  int temp_length;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;

  // multiplier=1.0;

  //  for(i=1;i<length;i++)
  //  multiplier=multiplier*i; // permutation factor
  // multiplier=multiplier*(length+2)*(length+1)*length; // factor from M()
  
  // (length-1)!        -- permutation factor as we sum over k
  // (length+1)*length  -- internal factor from H(). Note that 
  //                       summation over n.n. gives m thus 
  //                       length ->length+1.

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];

    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++)
	  conf[j][m]=configuration[j][m]; // M(k,m,m)
      
      temp_length=length+1;
      
      // we repeat the `m' 
      
      for(m=1;m<5;m++)
	conf[temp_length][m]=configuration[k][m];
            
      for(neighbour=1;neighbour<=hh[term].m_to_k_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* m ---> k */
	
	nn[1]=configuration[k][1]+
	  hh[term].m_to_k[index][neighbour][0];
	nn[2]=configuration[k][2]+
	  hh[term].m_to_k[index][neighbour][1];
	nn[3]=configuration[k][3]+
	  hh[term].m_to_k[index][neighbour][2];
	nn[4]=hh[term].m_to_k[index][neighbour][3];
	
	numberFlips1=1;
	
	for(l=1;l<=temp_length;l++)
	  if (nn[1]==conf[l][1] &&
	      nn[2]==conf[l][2] &&
	      nn[3]==conf[l][3] &&
	      nn[4]==conf[l][4])
	    numberFlips1++;
	
	numberFlips2=0;
	
	for(l=1;l<=temp_length;l++)
	  if (configuration[k][1]==conf[l][1] &&
	      configuration[k][2]==conf[l][2] &&
	      configuration[k][3]==conf[l][3] &&
	      configuration[k][4]==conf[l][4])
	    numberFlips2++;
	
	maxNumberFlips1=UnitCell.s2[nn[4]];
	maxNumberFlips2=UnitCell.s2[index];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips1 <= maxNumberFlips1 &&
	    numberFlips2 <= maxNumberFlips2) {	
	  
	  temp_length=length+2;
	  
	  conf[temp_length][1]=nn[1];
	  conf[temp_length][2]=nn[2];
	  conf[temp_length][3]=nn[3];
	  conf[temp_length][4]=nn[4];
	  
	  findDescriptorOfCluster(temp_length,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
	  
	  temp_multiplier=(-2)*0.5*UnitCell.s2[nn[4]];
	  
	  // coefficient from \tilde s^- \tilde s^- expansion
	  
//	  if (descriptorIndex > 0)
//	    sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
//		hh[term].coefficient[index][neighbour];	
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=37;
	      indices[2]=descriptorIndex;
              indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	    }
	  }

	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}


double term38(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 38: 

     N(k)(k)(m)(m) . S(+)(k) . S(+)(m).

  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index, index1, index2;
  int maxNumberFlips1, maxNumberFlips2;
  int numberFlips1, numberFlips2;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  // double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int temp_length;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));
  conf=         (int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<MAX_LSUBM+3;k++) {
    for(l=1;l<5;l++) {
      conf[k][l]=0;
    }
  } 
  
  // multiplier=1.0;
  
  //  for(i=1;i<=(length-2);i++)
  //  multiplier=multiplier*i; // permutation factor. 
  // multiplier=multiplier*(length+2)*(length+1)*length*(length-1);
  
  // (length-2)!      -- permutation factor because we sum over k1 and k2.
  // (length+2)*(length+1)*length*(length-1) 
  // .................-- factor from N()

  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {

	for(i=1;i<=length;i++)
	  for(j=1;j<5;j++)
	    conf[i][j]=configuration[i][j];
	
	// a check here...
	
	// now add in extra `k' and `m' sites for N(k,k,m,m)
	
	temp_length=length+1;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k1][m]; // `k'
	
	temp_length=length+2;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k2][m]; // `m'
	
	numberFlips1=1;
	numberFlips2=1;
	
	for(l=1;l<=temp_length-2;l++) {
	  if (configuration[k1][1]==conf[l][1] &&
	      configuration[k1][2]==conf[l][2] &&
	      configuration[k1][3]==conf[l][3] &&
	      configuration[k1][4]==conf[l][4])
	    numberFlips1++;
	
	  if (configuration[k2][1]==conf[l][1] &&
	      configuration[k2][2]==conf[l][2] &&
	      configuration[k2][3]==conf[l][3] &&
	      configuration[k2][4]==conf[l][4])
	    numberFlips2++;
	}

	index1=configuration[k1][4];
	maxNumberFlips1=UnitCell.s2[index1];
	
	index2=configuration[k2][4];
	maxNumberFlips2=UnitCell.s2[index2];

	are_they_neighbours=0;
	index=configuration[k1][4]; 
	
	for(j=1;j<=hh[term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  nn[1]=configuration[k1][1]+
	    hh[term].k_to_m[index][j][0];
	  nn[2]=configuration[k1][2]+
	    hh[term].k_to_m[index][j][1];
	  nn[3]=configuration[k1][3]+
	    hh[term].k_to_m[index][j][2];
	  nn[4]=hh[term].k_to_m[index][j][3];
	  
	  if(configuration[k2][1] == nn[1] &&
	     configuration[k2][2] == nn[2] &&
	     configuration[k2][3] == nn[3] &&
	     configuration[k2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}
	
	if ((maxNumberFlips1 > 1) && 
	    (maxNumberFlips2 > 1) &&
	    (numberFlips1 <= maxNumberFlips1) &&
	    (numberFlips2 <= maxNumberFlips2) &&
	    (are_they_neighbours)) {
	  
	  findDescriptorOfCluster(temp_length,
					conf,
					descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
//	  if (descriptorIndex > 0)
//	    sum+=CCMcoeffs[descriptorIndex]*
//	      hh[term].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=38;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
              indices[5]=0;
	      dummy=
		hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
	    }
	  }

	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

double term39(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 39: 
     
     F(m) . M(k)(k)(m) . S(+)(k) . S(+)(m)
     
    */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {	  

	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // M(k,k,m) k--1st one
	    conf1[2][m]=configuration[k1][m]; // M(k,k,m) k--2nd one
	    conf1[3][m]=configuration[k2][m]; // M(k,k,m) m 
	    conf2[1][m]=configuration[k2][m]; // F(m)     m
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[term].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+hh[term].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+hh[term].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+hh[term].k_to_m[index][j][2];
	    nn[4]=hh[term].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      
	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=3;
	      conf2_len=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H
		  conf1_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_len][m]=configuration[temp][m];
		} else {
		  // H
		  conf2_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_len][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }

	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		findDescriptorOfCluster(conf1_len,
					      conf1,
					      descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		findDescriptorOfCluster(conf2_len,
					      conf2,
					      descriptor);		
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex2 <= 0) continue;
		multiplier=1.0;

		/* for(i=1;i<=(conf1_len-3);i++)
		  multiplier=multiplier*i;
		multiplier=multiplier*conf1_len*
		  (conf1_len-1)*
		  (conf1_len-2);
		*/

		// M(k,k,m)             
		// (length-3)!                  -- because we sum over k and m
		// length*(length-1)*(length-2) -- factor due to M().
		
		// for(i=1;i<conf2_len;i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf2_len;
		
		// F(m)             
		// (length-1)!        -- because we sum over k and m
		// length             -- factor due to F().

		multiplier=2.0;

		// coefficient from \tilde s^- \tilde s^- expansion
		
//		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//		  sum+=multiplier
//		    *CCMcoeffs[descriptorIndex1]
//		    *CCMcoeffs[descriptorIndex2]*
//		    hh[term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=39;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term40(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 40:
     
     2 * F(k) . H(m)(m) . S(+)(m) . S(z)(k)
     
  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips1;
  int numberFlips2;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) { // `m' here. 
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];
    
    if (maxNumberFlips > 1) {
      
      for(m=1;m<5;m++) {
	conf1[1][m]=configuration[k][m]; // H(m,m) -- 1st m 
	conf1[2][m]=configuration[k][m]; // H(m,m) -- 2nd m
      }
      
      counter=1;
      for(m=1;m<=length;m++) {
	if (m != k) {
	  tempArray[counter]=m;
	  counter++;
	}
      }
      
      counter--;
      
      for(l=0;l<binary_max[counter];l++) {
	
	binary(l,counter,spin);
	conf1_len=2;
	conf2_len=0;
	
	for (j=1;j<=counter;j++) { 
	  if (spin[j]) {
	    // F(m)
	    conf1_len++;
	    temp=tempArray[j];
	    for (m=1;m<5;m++) 
	      conf1[conf1_len][m]=configuration[temp][m];
	  } else {
	    // H(k,m)
	    conf2_len++;
	    temp=tempArray[j];
	    for (m=1;m<5;m++) 
	      conf2[conf2_len][m]=configuration[temp][m];
	  }
	}
	
	index=conf1[1][4];
	
	for(neighbour=1;neighbour<=
	      hh[term].m_to_k_number[index];
	    neighbour++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  /* m ---> k */
	  
	  nn[1]=conf1[1][1]+
	    hh[term].m_to_k[index][neighbour][0];
	  nn[2]=conf1[1][2]+
	    hh[term].m_to_k[index][neighbour][1];
	  nn[3]=conf1[1][3]+
	    hh[term].m_to_k[index][neighbour][2];
	  nn[4]=hh[term].m_to_k[index][neighbour][3];
	  
	  numberFlips1=1;
	
	  for(m=1;m<=conf2_len;m++)
	    if (nn[1]==conf2[m][1] &&
		nn[2]==conf2[m][2] &&
		nn[3]==conf2[m][3] &&
		nn[4]==conf2[m][4])
	      numberFlips1++;
	
	  maxNumberFlips1=UnitCell.s2[nn[4]];
	
	  numberFlips2=0;
	  
	  for(m=1;m<=conf1_len;m++)
	    if (configuration[k][1]==conf1[m][1] &&
		configuration[k][2]==conf1[m][2] &&
		configuration[k][3]==conf1[m][3] &&
		configuration[k][4]==conf1[m][4])
	      numberFlips2++;
	
	  maxNumberFlips2=UnitCell.s2[configuration[k][4]];

	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	
	  if (numberFlips1 <= maxNumberFlips1 &&
	      numberFlips2 <= maxNumberFlips2) {
	  	  
	    temp_number_conf2_sites=conf2_len+1;
	  
	    conf2[temp_number_conf2_sites][1]=nn[1];
	    conf2[temp_number_conf2_sites][2]=nn[2];
	    conf2[temp_number_conf2_sites][3]=nn[3];
	    conf2[temp_number_conf2_sites][4]=nn[4];
	  
	    // H(m,m)
	    findDescriptorOfCluster(conf1_len,
					  conf1,
					  descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	
                if (descriptorIndex1 <= 0) continue;
	    // F(k)
	    findDescriptorOfCluster(temp_number_conf2_sites,
					  conf2,
					  descriptor);
	
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	
                if (descriptorIndex2 <= 0) continue;
	    multiplier=1.0;
	    /*	    for(i=1;i<=(conf1_len-2);i++)
	      multiplier=multiplier*i; 
	    multiplier=multiplier*conf1_len*
	      (conf1_len-1);
	    */

	    // H(m,m)
	    // (length-2)!       -- because we have m in H(m,m) twice
	    //                      and we sum over it.
	    // length*(length-1) -- internal factor in H(m,m)

	    // for(i=1;i<temp_number_conf2_sites;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*temp_number_conf2_sites;

	    // F(k)
	    // (length-1)!       -- permutation + we sum over k
	    // length            -- internal factor in F(k) 
	  
	    multiplier=(-2)*0.5*UnitCell.s2[nn[4]];

	    // factor from 2*S(z)(k)

//	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//	      sum+=multiplier
//		*CCMcoeffs[descriptorIndex1]
//		*CCMcoeffs[descriptorIndex2]*
//		hh[term].coefficient[index][neighbour];

	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=40;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*
		  hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term41(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 41: 
     
     F(k) . M(m)(m)(k) . S(+)(k) . S(+)(m)
     
    */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // M(m,m,k) m--1st one
	    conf1[2][m]=configuration[k1][m]; // M(m,m,k) m--2nd one
	    conf1[3][m]=configuration[k2][m]; // M(m,m,k) k 
	    conf2[1][m]=configuration[k2][m]; // F(k)     k
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[term].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+hh[term].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+hh[term].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+hh[term].m_to_k[index][j][2];
	    nn[4]=hh[term].m_to_k[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      
	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=3;
	      conf2_len=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H
		  conf1_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_len][m]=configuration[temp][m];
		} else {
		  // H
		  conf2_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_len][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		findDescriptorOfCluster(conf1_len,
					      conf1,
					      descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		findDescriptorOfCluster(conf2_len,
					      conf2,
					      descriptor);
		
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex2 <= 0) continue;
		multiplier=1.0;

		/* for(i=1;i<=(conf1_len-3);i++)
		  multiplier=multiplier*i;

		multiplier=multiplier*conf1_len*
		  (conf1_len-1)*
		  (conf1_len-2);
		*/

		// M(m,m,k)             
		// (length-3)!                   -- because we sum over k and m
		// length*(length-1)*(length-2)  -- factor due to M().
		
		//		for(i=1;i<conf2_len;i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf2_len;
		
		// F(k)             
		// (length-1)!        -- because we sum over k and m
		// length             -- factor due to F().

		multiplier=2.0; 

		// coefficient from \tilde s^- \tilde s^- expansion
		
//		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//		  sum+=multiplier
//		    *CCMcoeffs[descriptorIndex1]
//		    *CCMcoeffs[descriptorIndex2]*
//		    hh[term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=41;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term42(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 42:
     
     2 * F(m) . H(k)(k) . S(+)(k) . S(z)(m)
     
  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips1;
  int numberFlips2;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) { // `m' here. 
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];
    
    if (maxNumberFlips > 1) {
      
      for(m=1;m<5;m++) {
	conf1[1][m]=configuration[k][m]; // H(k,k) -- 1st k 
	conf1[2][m]=configuration[k][m]; // H(k,k) -- 2nd k
      }
      
      counter=1;
      for(m=1;m<=length;m++) {
	if (m != k) {
	  tempArray[counter]=m;
	  counter++;
	}
      }
      
      counter--;
      
      for(l=0;l<binary_max[counter];l++) {
	
	binary(l,counter,spin);
	conf1_len=2;
	conf2_len=0;
	
	for (j=1;j<=counter;j++) { 
	  if (spin[j]) {
	    // F(m)
	    conf1_len++;
	    temp=tempArray[j];
	    for (m=1;m<5;m++) 
	      conf1[conf1_len][m]=configuration[temp][m];
	  } else {
	    // H(k,m)
	    conf2_len++;
	    temp=tempArray[j];
	    for (m=1;m<5;m++) 
	      conf2[conf2_len][m]=configuration[temp][m];
	  }
	}
	
	index=conf1[1][4];
	
	for(neighbour=1;
	    neighbour<=hh[term].k_to_m_number[index];
	    neighbour++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  /* k ---> m */
	  
	  nn[1]=conf1[1][1]+
	    hh[term].k_to_m[index][neighbour][0];
	  nn[2]=conf1[1][2]+
	    hh[term].k_to_m[index][neighbour][1];
	  nn[3]=conf1[1][3]+
	    hh[term].k_to_m[index][neighbour][2];
	  nn[4]=hh[term].k_to_m[index][neighbour][3];
	  
	  numberFlips1=1;
	
	  for(m=1;m<=conf2_len;m++)
	    if (nn[1]==conf2[m][1] &&
		nn[2]==conf2[m][2] &&
		nn[3]==conf2[m][3] &&
		nn[4]==conf2[m][4])
	      numberFlips1++;
	
	  maxNumberFlips1=UnitCell.s2[nn[4]];
	
	  numberFlips2=0;
	  
	  for(m=1;m<=conf1_len;m++)
	    if (configuration[k][1]==conf1[m][1] &&
		configuration[k][2]==conf1[m][2] &&
		configuration[k][3]==conf1[m][3] &&
		configuration[k][4]==conf1[m][4])
	      numberFlips2++;
	
	  maxNumberFlips2=UnitCell.s2[configuration[k][4]];

	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	
	  if (numberFlips1 <= maxNumberFlips1 &&
	      numberFlips2 <= maxNumberFlips2) {
	  	  
	    temp_number_conf2_sites=conf2_len+1;
	  
	    conf2[temp_number_conf2_sites][1]=nn[1];
	    conf2[temp_number_conf2_sites][2]=nn[2];
	    conf2[temp_number_conf2_sites][3]=nn[3];
	    conf2[temp_number_conf2_sites][4]=nn[4];
	  
	    // H(k,k)
	    findDescriptorOfCluster(conf1_len,
					  conf1,
					  descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	
                if (descriptorIndex1 <= 0) continue;
	    // F(m)
	    findDescriptorOfCluster(temp_number_conf2_sites,
					  conf2,
					  descriptor);
	
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	
                if (descriptorIndex2 <= 0) continue;
	    multiplier=1.0;
	    /*	    for(i=1;i<=(conf1_len-2);i++)
	      multiplier=multiplier*i; 
	    multiplier=multiplier*(conf1_len-1)*
	      conf1_len;
	    */
	    // H(k,k)
	    // (length-2)!       -- because we have k in H(k,k) twice
	    //                      and we sum over it.
	    // length*(length-1) -- internal factor in H(k,k)

	    /*	    for(i=1;i<temp_number_conf2_sites;i++)
	      multiplier=multiplier*i;
	    multiplier=multiplier*temp_number_conf2_sites;
	    */

	    // F(m)
	    // (length-1)!       -- permutation factor from F(m)
	    // length            -- internal factor in F(m) 
	  
	    multiplier=(-2)*0.5*UnitCell.s2[nn[4]];

	    // factor from 2*S(z)(k)

//	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//	      sum+=multiplier
//		*CCMcoeffs[descriptorIndex1]
//		*CCMcoeffs[descriptorIndex2]*
//		hh[term].coefficient[index][neighbour];
	  
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=42;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*
		hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	      }
	    }

	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term43(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 43: 
     
     H(k)(k) . H(m)(m) . S(+)(k) . S(+)(m)
     
    */
  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);


  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  // double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips1;
  int numberFlips2;
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	maxNumberFlips2=UnitCell.s2[configuration[k2][4]];
	
	if (maxNumberFlips1 > 1 && maxNumberFlips2 > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // H(k,k) k--1st one
	    conf1[2][m]=configuration[k1][m]; // H(k,k) k--2nd one
	    conf2[1][m]=configuration[k2][m]; // H(m,m) m--1st one
	    conf2[2][m]=configuration[k2][m]; // H(m,m) m--2nd one
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[term].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+hh[term].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+hh[term].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+hh[term].k_to_m[index][j][2];
	    nn[4]=hh[term].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;
      
	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=2;
	      conf2_len=2;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H
		  conf1_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_len][m]=configuration[temp][m];
		} else {
		  // H
		  conf2_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_len][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];

	      numberFlips2=0;
	      
	      for(m=1;m<=conf2_len;m++) {
		if (configuration[k2][1]==conf2[m][1] &&
		    configuration[k2][2]==conf2[m][2] &&
		    configuration[k2][3]==conf2[m][3] &&
		    configuration[k2][4]==conf2[m][4])
		  numberFlips2++;
	      }
	      maxNumberFlips2=UnitCell.s2[configuration[k2][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1 &&
		  numberFlips2 <= maxNumberFlips2) {
		
		findDescriptorOfCluster(conf1_len,
					      conf1,
					      descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		findDescriptorOfCluster(conf2_len,
					      conf2,
					      descriptor);
		
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex2 <= 0) continue;
		// multiplier=1.0;
		/* for(i=1;i<=(conf1_len-2);i++)
		  multiplier=multiplier*i;
		multiplier=multiplier*conf1_len*
		  (conf1_len-1);
		*/

		// H(k,k)             
		// (length-2)!        -- because we sum over k and m
		// length*(length-1)  -- factor due to H().
		
		/*		for(i=1;i<=(conf2_len-2);i++)
		  multiplier=multiplier*i;
		multiplier=multiplier*conf2_len*
		  (conf2_len-1);
		*/
		// H(m,m)             
		// (length-2)!        -- because we sum over k and m
		// length*(length-1)  -- factor due to H().

//		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//		  sum+=CCMcoeffs[descriptorIndex1]
//		      *CCMcoeffs[descriptorIndex2]
//		      *hh[term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=43;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=
		      hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term44(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 44:
     
     H(k)(k) . F(m) . F(m) . S(+)(k) . S(+)(m)
     
  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int mm,nnn,temp1,temp2,temp3;
  int k1, k2;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  // double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5], (*conf3)[5];
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int are_they_neighbours;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
  
  for(k1=1;k1<=length;k1++) { 
    for(k2=1;k2<=length;k2++) { 
      
      if (k1 != k2) {
	
	index=configuration[k1][4];
	maxNumberFlips=UnitCell.s2[index];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // H(k,k) <-- 1st k here
	    conf1[2][m]=configuration[k1][m]; // H(k,k) <-- 2nd k here
	    conf2[1][m]=configuration[k2][m]; // F(m)
	    conf3[1][m]=configuration[k2][m]; // F(m)
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[term].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+hh[term].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+hh[term].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+hh[term].k_to_m[index][j][2];
	    nn[4]=hh[term].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      

	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(m=1;m<=length;m++) {
	      if (m != k1 && m != k2) {
		tempArray[counter]=m;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    if (counter > 30) {
	      printf("TERM 44: counter > 30...\nExiting!!\n\n");
	      exit(1);
	    }
            temp = counter; // split into summands not greater than 10 each
            temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
            temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
            temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	      
	    for(l=0;l<trimary_max[temp1];l++) {
	      for(mm=0;mm<trimary_max[temp2];mm++) {
		for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		    
		  trimary(l,mm,nnn,counter,spin);
		    
		  conf1_len=2;
		  conf2_len=1;
		  conf3_len=1;
		    
		  for (j=1;j<=counter;j++) { 
		    if (spin[j]==0) {
		      // H(k,k)
		      conf1_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf1[conf1_len][m]=configuration[temp][m];
		    } else if (spin[j]==1) {
		      // F(m)
		      conf2_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf2[conf2_len][m]=configuration[temp][m];
		    } else {
		      // F(m)
		      conf3_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf3[conf3_len][m]=configuration[temp][m];
		    }
		  }
	      
		  numberFlips=0;
	      
		  for(m=1;m<=conf1_len;m++)
		    if (configuration[k1][1]==conf1[m][1] &&
			configuration[k1][2]==conf1[m][2] &&
			configuration[k1][3]==conf1[m][3] &&
			configuration[k1][4]==conf1[m][4])
		      numberFlips++;
	      
		  maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	      
		  // the number of spin flips that you can have at a given
		  // lattice site is depedent on the quantum spin number
		  // of the site in the UNIT CELL index `length'
	      
		  if (numberFlips <= maxNumberFlips) {
		
		    // H(k,k)
		    findDescriptorOfCluster(conf1_len,
						  conf1,
						  descriptor);
		    descriptorIndex1=findDescriptorIndex(descriptor,
							 NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		    // F(m)
		    findDescriptorOfCluster(conf2_len,
						  conf2,
						  descriptor);
		
		    descriptorIndex2=findDescriptorIndex(descriptor,
							 NumberConfigurations);
	
		
                if (descriptorIndex2 <= 0) continue;
		    // F(m)
		    findDescriptorOfCluster(conf3_len,
						  conf3,
						  descriptor);
		
		    descriptorIndex3=findDescriptorIndex(descriptor,
							 NumberConfigurations);
		
		
                if (descriptorIndex3 <= 0) continue;
		    // multiplier=1.0;
		    /* for(i=1;i<=conf1_len-2;i++)
		      multiplier=multiplier*i; 
		    multiplier=multiplier*(conf1_len-1)*
		      conf1_len;
		    */
		    // H(k,k)
		    // (length-2)!       -- because we have k in H(k,k) twice
		    //                      and we sum over it.
		    // length(length-1)  -- internal factor in H(k,k)
		    /*
		    for(i=1;i<conf2_len;i++)
		      multiplier=multiplier*i;
		    multiplier=multiplier*conf2_len;
		    */
		    // F(m)
		    // (length-1)!       -- permutation factor from F(m)
		    // length            -- internal factor in F(m) 
		
		    /*		    
                       for(i=1;i<conf3_len;i++)
		         multiplier=multiplier*i;
		       multiplier=multiplier*conf3_len;
		    */
		    // F(m)
		    // (length-1)!       -- permutation factor from F(m)
		    // length            -- internal factor in F(m) 
		
//		    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
//			(descriptorIndex3 > 0))
//		      sum+=CCMcoeffs[descriptorIndex1]
//			  *CCMcoeffs[descriptorIndex2]
//			  *CCMcoeffs[descriptorIndex3]
//			  *hh[term].coefficient[index][are_they_neighbours];
	  
		    if (Control.file) {
		      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
			  (descriptorIndex3 > 0)) {
			int indices[7]={0};
			double dummy;
			indices[0]=config_index;
			indices[1]=44;
			indices[2]=descriptorIndex1;
			indices[3]=descriptorIndex2;
			indices[4]=descriptorIndex3;
			indices[5]=0;
			dummy=
			  hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
		      }
		    }

		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}  

double term45(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 45:
     
     H(m)(m) . F(k) . F(k) . S(+)(k) . S(+)(m)
     
  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int mm,nnn,temp1,temp2,temp3;
  int k1, k2;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  // double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5], (*conf3)[5];
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int are_they_neighbours;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 

  for(k1=1;k1<=length;k1++) { 
    for(k2=1;k2<=length;k2++) { 
      
      if (k1 != k2) {
	
	index=configuration[k1][4];
	maxNumberFlips=UnitCell.s2[index];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // H(m,m) <-- 1st m here
	    conf1[2][m]=configuration[k1][m]; // H(m,m) <-- 2nd m here
	    conf2[1][m]=configuration[k2][m]; // F(k)
	    conf3[1][m]=configuration[k2][m]; // F(k)
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[term].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+hh[term].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+hh[term].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+hh[term].m_to_k[index][j][2];
	    nn[4]=hh[term].m_to_k[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      

	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(m=1;m<=length;m++) {
	      if (m != k1 && m != k2) {
		tempArray[counter]=m;
		counter++;
	      }
	    }
	    
	    counter--;

	    if (counter > 30) {
	      printf("TERM 45: counter > 30...\nExiting!!\n\n");
	      exit(1);
	    }
            temp = counter; // split into summands not greater than 10 each
            temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
            temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
            temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	    
	    for(l=0;l<trimary_max[temp1];l++) {
	      for(mm=0;mm<trimary_max[temp2];mm++) {
		for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		  
		  trimary(l,mm,nnn,counter,spin);
		  
		  conf1_len=2;
		  conf2_len=1;
		  conf3_len=1;

		  for (j=1;j<=counter;j++) { 
		    if (spin[j]==0) {
		      // H(m,m)
		      conf1_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf1[conf1_len][m]=configuration[temp][m];
		    } else if (spin[j]==1) {
		      // F(k)
		      conf2_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf2[conf2_len][m]=configuration[temp][m];
		    } else {
		      // F(k)
		      conf3_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf3[conf3_len][m]=configuration[temp][m];
		    }
		  }
	      
		  numberFlips=0;
	      
		  for(m=1;m<=conf1_len;m++)
		    if (configuration[k1][1]==conf1[m][1] &&
			configuration[k1][2]==conf1[m][2] &&
			configuration[k1][3]==conf1[m][3] &&
			configuration[k1][4]==conf1[m][4])
		      numberFlips++;
	      
		  maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	      
		  // the number of spin flips that you can have at a given
		  // lattice site is depedent on the quantum spin number
		  // of the site in the UNIT CELL index `length'
	      
		  if (numberFlips <= maxNumberFlips) {
		
		    // H(m,m)
		    findDescriptorOfCluster(conf1_len,
						  conf1,
						  descriptor);
		    descriptorIndex1=findDescriptorIndex(descriptor,
							 NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		    // F(k)
		    findDescriptorOfCluster(conf2_len,
						  conf2,
						  descriptor);
		
		    descriptorIndex2=findDescriptorIndex(descriptor,
							 NumberConfigurations);
	
                if (descriptorIndex2 <= 0) continue;
		
		    // F(k)
		    findDescriptorOfCluster(conf3_len,
						  conf3,
						  descriptor);
		
		    descriptorIndex3=findDescriptorIndex(descriptor,
							 NumberConfigurations);
		
                if (descriptorIndex3 <= 0) continue;
		
		    // multiplier=1.0;
		    /* for(i=1;i<=(conf1_len-2);i++)
		      multiplier=multiplier*i; 
		    multiplier=multiplier*(conf1_len-1)*
		      conf1_len;
		    */
		    // H(m,m)
		    // (length-2)!       -- because we have m in H(m,m) twice
		    //                      and we sum over it.
		    // length(length-1)  -- internal factor in H(m,m)
		    /*
		    for(i=1;i<conf2_len;i++)
		      multiplier=multiplier*i;
		    multiplier=multiplier*conf2_len;
		    */
		    // F(k)
		    // (length-1)!       -- permutation factor from F(k)
		    // length            -- internal factor in F(k) 
		
		    /*		    for(i=1;i<conf3_len;i++)
		      multiplier=multiplier*i;
		    multiplier=multiplier*conf3_len;
		    */
		    // F(k)
		    // (length-1)!       -- permutation factor from F(k)
		    // length            -- internal factor in F(k) 
		
//		    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
//			(descriptorIndex3 > 0))
//		      sum+=CCMcoeffs[descriptorIndex1]
//			*CCMcoeffs[descriptorIndex2]
//			*CCMcoeffs[descriptorIndex3];

		    if (Control.file) {
		      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
			  (descriptorIndex3 > 0)) {
			int indices[7]={0};
			double dummy;
			indices[0]=config_index;
			indices[1]=45;
			indices[2]=descriptorIndex1;
			indices[3]=descriptorIndex2;
			indices[4]=descriptorIndex3;
			indices[5]=0;
			dummy=hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
		      }
		    }

		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}  

double term46(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 46:
     
     - H(m)(m) . S(+)(m) . S(z)(k)
     
  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips2;
  int numberFlips2;
  int temp_length;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;

  multiplier=1.0;
  // for(i=1;i<length;i++)
  //  multiplier=multiplier*i; // permutation factor
  // multiplier=multiplier*(length+1)*length; // factor from M()
  
  // (length-1)!        -- permutation factor as we sum over k
  // (length+1)*length  -- internal factor from H(). Note that 
  //                       summation over n.n. gives m thus 
  //                       length ->length+1.

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];

    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++)
	  conf[j][m]=configuration[j][m]; // H(m,m) <- 1st m
      
      temp_length=length+1;
      
      // we repeat the `m' 
      
      for(m=1;m<5;m++)
	conf[temp_length][m]=configuration[k][m]; // H(m,m) <- 2nd m

      index=configuration[k][4];
      
      for(neighbour=1;neighbour<=
	    hh[term].m_to_k_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* m ---> k */
	
	nn[1]=configuration[k][1]+
	  hh[term].m_to_k[index][neighbour][0];
	nn[2]=configuration[k][2]+
	  hh[term].m_to_k[index][neighbour][1];
	nn[3]=configuration[k][3]+
	  hh[term].m_to_k[index][neighbour][2];
	nn[4]=hh[term].m_to_k[index][neighbour][3];
	
	numberFlips2=0;
	
	for(l=1;l<=temp_length;l++)
	  if (configuration[k][1]==conf[l][1] &&
	      configuration[k][2]==conf[l][2] &&
	      configuration[k][3]==conf[l][3] &&
	      configuration[k][4]==conf[l][4])
	    numberFlips2++;
	
	maxNumberFlips2=UnitCell.s2[index];

	if (numberFlips2 <= maxNumberFlips2)  {
	  
	  findDescriptorOfCluster(temp_length,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  temp_multiplier=multiplier*0.5*UnitCell.s2[nn[4]];
	  
	  // coefficient from \tilde s^z \tilde s^- expansion
	  
//	  if (descriptorIndex > 0)
//	    sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
//	      hh[term].coefficient[index][neighbour];	
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=46;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	    }
	  }
	  
	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

double term47(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 47: 

     - M(k)(m)(m) . S(+)(k) . S(+)(m).

  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index, index1;
  int maxNumberFlips1;
  int numberFlips1;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  // double multiplier;
  double temp_multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int temp_length;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<MAX_LSUBM+3;k++) 
    for(l=1;l<5;l++) 
      conf[k][l]=0;
    
   
  // multiplier=1.0;
  
  /*  for(i=1;i<=(length-2);i++)
    multiplier=multiplier*i; 
  multiplier=multiplier*(length+1)*length*(length-1);
  */

  // (length-2)!      -- permutation factor because we sum over k1 and k2.
  // (length+1)length(length-1) -- factor from M()

  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {

	for(i=1;i<=length;i++)
	  for(j=1;j<5;j++)
	    conf[i][j]=configuration[i][j];
	
	// a check here...
	
	// now add in extra `m' site for M(k,m,m)
	
	temp_length=length+1;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k1][m]; // `m'
	
	numberFlips1=0;
	
	for(l=1;l<=temp_length;l++) {
	  if (configuration[k1][1]==conf[l][1] &&
	      configuration[k1][2]==conf[l][2] &&
	      configuration[k1][3]==conf[l][3] &&
	      configuration[k1][4]==conf[l][4])
	    numberFlips1++;
	
	}

	index1=configuration[k1][4];
	maxNumberFlips1=UnitCell.s2[index1];

	are_they_neighbours=0;
	index=configuration[k1][4]; 
	
	for(j=1;j<=hh[term].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  // m -- > k

	  nn[1]=configuration[k1][1]+
	    hh[term].m_to_k[index][j][0];
	  nn[2]=configuration[k1][2]+
	    hh[term].m_to_k[index][j][1];
	  nn[3]=configuration[k1][3]+
	    hh[term].m_to_k[index][j][2];
	  nn[4]=hh[term].m_to_k[index][j][3];
	  
	  if(configuration[k2][1] == nn[1] &&
	     configuration[k2][2] == nn[2] &&
	     configuration[k2][3] == nn[3] &&
	     configuration[k2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}
	
	if ((maxNumberFlips1 > 1) && 
	    (numberFlips1 <= maxNumberFlips1) &&
	    (are_they_neighbours)) {
	  
	  
	  findDescriptorOfCluster(temp_length,
					conf,
					descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  temp_multiplier=-1.0;

//	  if (descriptorIndex > 0)
//	    sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
//	      hh[term].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=47;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
	    }
	  }

	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

double term48(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 48: 
     
     - F(k) . H(m)(m) . S(+)(k) . S(+)(m)
     
    */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // H(m,m) m--1st one
	    conf1[2][m]=configuration[k1][m]; // H(m,m) m--2nd one
	    conf2[1][m]=configuration[k2][m]; // F(k)   <- k
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[term].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+hh[term].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+hh[term].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+hh[term].m_to_k[index][j][2];
	    nn[4]=hh[term].m_to_k[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;
  
	    if (are_they_neighbours) break;    
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=2;
	      conf2_len=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H
		  conf1_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_len][m]=configuration[temp][m];
		} else {
		  // H
		  conf2_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_len][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		findDescriptorOfCluster(conf1_len,
					      conf1,
					      descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		findDescriptorOfCluster(conf2_len,
					      conf2,
					      descriptor);
		
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex2 <= 0) continue;
		multiplier=1.0;
		/* for(i=1;i<=(conf1_len-2);i++)
		  multiplier=multiplier*i;
		multiplier=multiplier*conf1_len*
		  (conf1_len-1);
		*/
		// H(m,m)             
		// (length-2)!        -- because we sum over m
		// length*(length-1)  -- factor due to M().
		/*
		for(i=1;i<conf2_len;i++)
		  multiplier=multiplier*i;
		multiplier=multiplier*conf2_len;
		*/
		// F(k)             
		// (length-1)!        -- because we sum over k
		// length             -- factor due to F().

		multiplier=-1.0; 

		// coefficient from \tilde s^- \tilde s^- expansion
		
//		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//		  sum+=multiplier
//		    *CCMcoeffs[descriptorIndex1]
//		    *CCMcoeffs[descriptorIndex2]*
//		    hh[term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=48;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term49(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 49: 

     - M(k)(k)(m) . S(+)(k) . S(+)(m).

  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index, index1;
  int maxNumberFlips1;
  int numberFlips1;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  // double multiplier;
  double temp_multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int temp_length;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<MAX_LSUBM+3;k++) 
    for(l=1;l<5;l++) 
      conf[k][l]=0;
    
   
  // multiplier=1.0;
  
  //  for(i=1;i<=(length-2);i++)
  //  multiplier=multiplier*i; // permutation factor. 
  // multiplier=multiplier*(length+1)*length*(length-1);
  
  // (length-2)!      -- permutation factor because we sum over k1 and k2.
  // (length+1)*length*(length-1) -- factor from M()

  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {

	for(i=1;i<=length;i++)
	  for(j=1;j<5;j++)
	    conf[i][j]=configuration[i][j];
	
	// a check here...
	
	// now add in extra `k' site for M(k,k,m)
	
	temp_length=length+1;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k1][m]; // `k'
	
	numberFlips1=0;
	
	for(l=1;l<=temp_length;l++) {
	  if (configuration[k1][1]==conf[l][1] &&
	      configuration[k1][2]==conf[l][2] &&
	      configuration[k1][3]==conf[l][3] &&
	      configuration[k1][4]==conf[l][4])
	    numberFlips1++;
	}

	index1=configuration[k1][4];
	maxNumberFlips1=UnitCell.s2[index1];

	are_they_neighbours=0;
	index=configuration[k1][4]; 
	
	for(j=1;j<=hh[term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  // k -- > m

	  nn[1]=configuration[k1][1]+
	    hh[term].k_to_m[index][j][0];
	  nn[2]=configuration[k1][2]+
	    hh[term].k_to_m[index][j][1];
	  nn[3]=configuration[k1][3]+
	    hh[term].k_to_m[index][j][2];
	  nn[4]=hh[term].k_to_m[index][j][3];
	  
	  if(configuration[k2][1] == nn[1] &&
	     configuration[k2][2] == nn[2] &&
	     configuration[k2][3] == nn[3] &&
	     configuration[k2][4] == nn[4])
	    are_they_neighbours=j;   

	  if (are_they_neighbours) break; 
	}
	
	if ((maxNumberFlips1 > 1) && 
	    (numberFlips1 <= maxNumberFlips1) &&
	    (are_they_neighbours)) {
	  
	  
	  findDescriptorOfCluster(temp_length,
					conf,
					descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	 
	  temp_multiplier=-1.0;

//	  if (descriptorIndex > 0)
//	    sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
//	      hh[term].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=49;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
	    }
	  }

	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

double term50(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 50: (Term_13=5)
     
     - H(k)(k) . S(+)(k) . S(z)(m)
     
  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips2;
  int numberFlips2;
  int temp_length;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;

  multiplier=1.0;

  //  for(i=1;i<length;i++)
  //  multiplier=multiplier*i; // permutation factor
  // multiplier=multiplier*(length+1)*length; // factor from M()
  
  // (length-1)!        -- permutation factor as we sum over k
  // (length+1)*length  -- internal factor from H(). Note that 
  //                       summation over n.n. gives m thus 
  //                       length ->length+1.

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];

    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++)
	  conf[j][m]=configuration[j][m]; 
      
      temp_length=length+1;
      
      // we repeat the `m' 
      
      for(m=1;m<5;m++)
	conf[temp_length][m]=configuration[k][m]; // H(k,k) <- 2nd k

      index=configuration[k][4];
      
      for(neighbour=1;neighbour<=hh[term].k_to_m_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=configuration[k][1]+
	  hh[term].k_to_m[index][neighbour][0];
	nn[2]=configuration[k][2]+
	  hh[term].k_to_m[index][neighbour][1];
	nn[3]=configuration[k][3]+
	  hh[term].k_to_m[index][neighbour][2];
	nn[4]=hh[term].k_to_m[index][neighbour][3];
	
	numberFlips2=0;
	
	for(l=1;l<=temp_length;l++)
	  if (configuration[k][1]==conf[l][1] &&
	      configuration[k][2]==conf[l][2] &&
	      configuration[k][3]==conf[l][3] &&
	      configuration[k][4]==conf[l][4])
	    numberFlips2++;
	
	maxNumberFlips2=UnitCell.s2[index];

	if (numberFlips2 <= maxNumberFlips2)  {
	  
	  findDescriptorOfCluster(temp_length,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  temp_multiplier=multiplier*0.5*UnitCell.s2[nn[4]];
          // * Hamiltonian (buggy, removed by DF Apr08)
	  
	  // coefficient from \tilde s^z \tilde s^- expansion
	  
//	  if (descriptorIndex > 0)
//	    sum+=temp_multiplier*CCMcoeffs[descriptorIndex];	
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=50;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		hh[term].coefficient[index][neighbour];
            collect_term(indices,dummy,h_id);
	    }
	  }
	  
	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

double term51(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 51: 
     
     - F(m) . H(k)(k) . S(+)(k) . S(+)(m)
     
    */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // H(k,k) k -- 1st one
	    conf1[2][m]=configuration[k1][m]; // H(k,k) k -- 2nd one
	    conf2[1][m]=configuration[k2][m]; // F(m)   m
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[term].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+hh[term].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+hh[term].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+hh[term].k_to_m[index][j][2];
	    nn[4]=hh[term].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;   

	    if (are_they_neighbours) break;   
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=2;
	      conf2_len=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H(k,k)
		  conf1_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_len][m]=configuration[temp][m];
		} else {
		  // F(m)
		  conf2_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_len][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		findDescriptorOfCluster(conf1_len,
					      conf1,
					      descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		findDescriptorOfCluster(conf2_len,
					      conf2,
					      descriptor);
		
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex2 <= 0) continue;
		multiplier=1.0;

		/*		for(i=1;i<=a(conf1_len-2);i++)
		  multiplier=multiplier*i;
		multiplier=multiplier*conf1_len*
		  (conf1_len-1);
		*/
		// H(k,k)             
		// (length-2)!        -- because we sum over k
		// length*(length-1)  -- factor due to M().
		/*
		for(i=1;i<conf2_len;i++)
		  multiplier=multiplier*i;
		multiplier=multiplier*conf2_len;
		*/

		// F(m)             
		// (length-1)!        -- because we sum over k and m
		// length             -- factor due to F().

		multiplier=-1.0; 

		// coefficient from \tilde s^- \tilde s^z expansion
		
//		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//		  sum+=multiplier
//		    *CCMcoeffs[descriptorIndex1]
//		    *CCMcoeffs[descriptorIndex2]*
//		    hh[term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=51;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term52(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 52: 

     - H(m)(m) . S(+)(k) . S(+)(m).

  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i, j, k, l, m;
  int k1;
  int k2;
  int counter;
  int nn[5];
  int are_they_neighbours;
  int index;
  int maxNumberFlips1;
  int numberFlips1;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int temp_length;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<MAX_LSUBM+3;k++) 
    for(l=1;l<5;l++) 
      conf[k][l]=0;
       
  multiplier=-1.0;
  
  // for(i=1;i<=(length-2);i++)
  //  multiplier=multiplier*i; // permutation factor. 
  // multiplier=multiplier*length*(length-1);
  
  // (length-2)!      -- permutation factor because we sum over k1 and k2.
  // length(length-1) -- factor from H()

  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {

	counter=1;
	for(i=1;i<=length;i++) {
	  if (i != k2) { 
	    for(j=1;j<5;j++)
	      conf[counter][j]=configuration[i][j];
	    counter++;
	  }
	}

	if (counter > length) {
	  printf("Error in term 52...\nExiting!!\n\n");
	  exit(1);
	}

	// a check here...
	
	// now add in extra `m' site for H(m,m)
	
	temp_length=length;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k1][m]; // `m'
	
	numberFlips1=0;
	
	for(l=1;l<=temp_length;l++) {
	  if (configuration[k1][1]==conf[l][1] &&
	      configuration[k1][2]==conf[l][2] &&
	      configuration[k1][3]==conf[l][3] &&
	      configuration[k1][4]==conf[l][4])
	    numberFlips1++;
	}

	are_they_neighbours=0;
	index=configuration[k1][4]; 
	maxNumberFlips1=UnitCell.s2[index];
	
	for(j=1;j<=hh[term].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  // m -- > k

	  nn[1]=configuration[k1][1]+
	    hh[term].m_to_k[index][j][0];
	  nn[2]=configuration[k1][2]+
	    hh[term].m_to_k[index][j][1];
	  nn[3]=configuration[k1][3]+
	    hh[term].m_to_k[index][j][2];
	  nn[4]=hh[term].m_to_k[index][j][3];
	  
	  if(configuration[k2][1] == nn[1] &&
	     configuration[k2][2] == nn[2] &&
	     configuration[k2][3] == nn[3] &&
	     configuration[k2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}
	
	if ((maxNumberFlips1 > 1) && 
	    (numberFlips1 <= maxNumberFlips1) &&
	    (are_they_neighbours)) {	  
	  
	  findDescriptorOfCluster(temp_length,
					conf,
					descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
//	  if (descriptorIndex > 0)
//	    sum+=multiplier*CCMcoeffs[descriptorIndex]*
//		hh[term].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=52;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*
		hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
	    }
	  }

	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

double term53(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 53: 

     - H(k)(k) . S(+)(k) . S(+)(m).

  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i, j, k, l, m;
  int k1;
  int k2;
  int counter;
  int nn[5];
  int are_they_neighbours;
  int index;
  int maxNumberFlips1;
  int numberFlips1;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int temp_length;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<MAX_LSUBM+3;k++) 
    for(l=1;l<5;l++) 
      conf[k][l]=0;
    
   
  multiplier=-1.0; // note the minus here!!
  
  //  for(i=1;i<=(length-2);i++)
  //  multiplier=multiplier*i; // permutation factor. 
  // multiplier=multiplier*length*(length-1);
  
  // (length-2)!      -- permutation factor because we sum over k1 and k2.
  // (length+1)length(length-1) -- factor from H()

  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {

	counter=1;
	for(i=1;i<=length;i++) {
	  if (i != k2) { 
	    for(j=1;j<5;j++)
	      conf[counter][j]=configuration[i][j];
	    counter++;
	  }
	}
	
	if (counter > length) {
	  printf("Error in term 53...\nExiting!!\n\n");
	  exit(1);
	}

	// a check here...
	
	// now add in extra `k' site for H(k,k)
	
	temp_length=length;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k1][m]; // `k'
	
	numberFlips1=0;
	
	for(l=1;l<=temp_length;l++) {
	  if (configuration[k1][1]==conf[l][1] &&
	      configuration[k1][2]==conf[l][2] &&
	      configuration[k1][3]==conf[l][3] &&
	      configuration[k1][4]==conf[l][4])
	    numberFlips1++;
	
	}

	are_they_neighbours=0;
	index=configuration[k1][4]; 
	maxNumberFlips1=UnitCell.s2[index];

	for(j=1;j<=hh[term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  // k --> m

	  nn[1]=configuration[k1][1]+
	    hh[term].k_to_m[index][j][0];
	  nn[2]=configuration[k1][2]+
	    hh[term].k_to_m[index][j][1];
	  nn[3]=configuration[k1][3]+
	    hh[term].k_to_m[index][j][2];
	  nn[4]=hh[term].k_to_m[index][j][3];
	  
	  if(configuration[k2][1] == nn[1] &&
	     configuration[k2][2] == nn[2] &&
	     configuration[k2][3] == nn[3] &&
	     configuration[k2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}

	
	if ((maxNumberFlips1 > 1) && 
	    (numberFlips1 <= maxNumberFlips1) &&
	    (are_they_neighbours)) {
	  
	  
	  findDescriptorOfCluster(temp_length,
					conf,
					descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
//	  if (descriptorIndex > 0)
//	    sum+=multiplier*CCMcoeffs[descriptorIndex]*
//		hh[term].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
              indices[1]=53;
	      indices[2]=descriptorIndex;
              indices[3]=0;
	      indices[4]=0;
              indices[5]=0;
	      dummy=multiplier*
		hh[term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,h_id);
	    }
	  }

	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

// Single-site terms now: s^z

double term54(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  
  /* 
     TERM 54: 
     
     F(k) . S(+)(k) 

     Wed Jul 10 BST 2002

  */
  
  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  int (*configuration)[5];
  int index;
  double sum;
  sum=0.0000;
  
  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];
  
  // get the multiplying coefficient correct first!

  for(k=1;k<=length;k++) {
    
    findDescriptorOfCluster(length,configuration,descriptor);
    descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
    
    if (descriptorIndex != config_index) {
      printf("Term 54 Error: descriptorIndex (%d"
	") != config_index (%d)...\nExiting!!\n\n",
        descriptorIndex,config_index);
      exit(1);
    }

    index=configuration[k][4];      
    
//    sum+=CCMcoeffs[config_index]*
//      hh[term].coefficient[index][1];
    
    if (Control.file) {
      int indices[7]={0};
      double dummy;
      indices[0]=config_index;
      indices[1]=54;
      indices[2]=config_index;
      indices[3]=0;
      indices[4]=0;
      indices[5]=0;
      dummy=hh[term].coefficient[index][1];
      collect_term(indices,dummy,h_id);
    }
  
  }
  
  free(configuration);
  return sum;
}

// Single-site: s^-

double term55(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 55: 
     
     -2 * F(k) . S(z)(k) 
  
  */
  
  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);
  // Oct05: configs with length+1 does not generate terms, so we skip it
  if (length+1>Params.n) return 0;

  int i, j, l;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  // double multiplier;
  // double temp_multiplier;
  double sum;
  sum=0.0000;
  
  int minX, minY, minZ;  
  int maxX, maxY, maxZ;
  
  int x,y,z,cell;
  int index;
  
  int numberFlips, maxNumberFlips;
  
  int (*configuration)[5];
  
  configuration=(int (*)[5])malloc((MAX_LSUBM+4)*5*sizeof(int));
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];
  
  int maxPointX;
  int maxPointY;
  int maxPointZ;
  
  maxPointX=0;
  maxPointY=0;
  maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1])) maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2])) maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3])) maxPointZ=abs(string[j][3]);
  }
  
  if (Params.dim == 1) {
    minX=radius;           minY=0;                minZ=0;
    maxX=radius+maxPointX; maxY=0;                maxZ=0;
  } else if (Params.dim == 2) {
    minX=radius;           minY=radius;           minZ=0;
    maxX=radius+maxPointX; maxY=radius+maxPointY; maxZ=0;
  } else if (Params.dim == 3) {
    minX=radius;           minY=radius;           minZ=radius;
    maxX=radius+maxPointX; maxY=radius+maxPointY; maxZ=radius+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d" 
      ") not recognised...\nExiting||\n\n",Params.dim);
    exit(1);
  }
  
  for(x=-minX;x<=maxX;x++) {
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {
	  
	  // order=length+1;   // this is the `k'    
	  
	  configuration[length+1][1]=x;
	  configuration[length+1][2]=y;
	  configuration[length+1][3]=z;
	  configuration[length+1][4]=cell;
	  
	  index=cell;
	  numberFlips=1;
	  
	  // check that we don't have too many spins at site `k'.
	  
	  for(l=1;l<=length;l++)
	    if (configuration[length+1][1]==configuration[l][1] &&
		configuration[length+1][2]==configuration[l][2] &&
		configuration[length+1][3]==configuration[l][3] &&
		configuration[length+1][4]==configuration[l][4])
	      numberFlips++;
	  
	  maxNumberFlips=UnitCell.s2[configuration[length+1][4]];
	  
	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	  
	  if (numberFlips <= maxNumberFlips) {
	    
	    // temp_multiplier=0.5*UnitCell.s2[configuration[length+1][4]]*
	    //  multiplier;
	    
	    findDescriptorOfCluster(length+1,
					  configuration,
					  descriptor);
	    
	    descriptorIndex=findDescriptorIndex(descriptor,
						NumberConfigurations);
	    
//	    if (descriptorIndex > 0)
//	      sum+=(-2.0)*(-0.5*UnitCell.s2[configuration[length+1][4]])
//		*CCMcoeffs[descriptorIndex]
//		*hh[term].coefficient[index][1];
	    
	    if (Control.file) {
	      if (descriptorIndex > 0){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=55;
		indices[2]=descriptorIndex;
		indices[3]=0;
		indices[4]=0;
		indices[5]=0;
		dummy=(-2.0)*(-0.5*UnitCell.s2[configuration[length+1][4]])*
		  hh[term].coefficient[index][1];
		collect_term(indices,dummy,h_id);
	      }
	    }
	    
	  }
	}
      }
    }    
  }
  return sum;  
}

double term56(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 56:
     
     - H(k)(k) . S(+)(k) 
  
  */
  
  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  // double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int index;
  int maxNumberFlips;
  int maxNumberFlips2;
  int numberFlips2;
  int temp_length;
  double sum;
  sum=0.0000;
  
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;
  
  // multiplier=1.0;
  
  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];
    
    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++)
	  conf[j][m]=configuration[j][m]; // H(m,m) <- 1st m
      
      temp_length=length+1;
      
      // we repeat the `m' 
      
      for(m=1;m<5;m++)
	conf[temp_length][m]=configuration[k][m]; // H(m,m) <- 2nd m
      
      index=configuration[k][4];
      
      numberFlips2=0;
      
      for(l=1;l<=temp_length;l++)
	if (configuration[k][1]==conf[l][1] &&
	    configuration[k][2]==conf[l][2] &&
	    configuration[k][3]==conf[l][3] &&
	    configuration[k][4]==conf[l][4])
	  numberFlips2++;
      
      maxNumberFlips2=UnitCell.s2[index];
      
      if (numberFlips2 <= maxNumberFlips2)  {
	
	findDescriptorOfCluster(temp_length,conf,descriptor);
	descriptorIndex=findDescriptorIndex(descriptor,
					    NumberConfigurations);
	
//	if (descriptorIndex > 0)
//	  sum+=(-1.0)*CCMcoeffs[descriptorIndex]*
//	    hh[term].coefficient[index][1];	
	
	if (Control.file) {
	  if (descriptorIndex > 0) {
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=56;
	    indices[2]=descriptorIndex;
	    indices[3]=0;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=(-1.0)*hh[term].coefficient[index][1];
            collect_term(indices,dummy,h_id);
	  }
	}
	
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

double term57(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 57: 
     
     - F(k) . F(k) . S(+)(k) 

  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int (*configuration)[5];
  int index;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];
  
  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // F(m)
    
    for(m=1;m<5;m++)
      conf2[1][m]=configuration[k][m]; // F(m)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(m) 1st one
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // F(m) 2nd one
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];	

      findDescriptorOfCluster(conf1_len,
				    conf1,
				    descriptor);
      descriptorIndex1=findDescriptorIndex(descriptor,
					   NumberConfigurations);
      
                if (descriptorIndex1 <= 0) continue;
      findDescriptorOfCluster(conf2_len,
				    conf2,
				    descriptor);
      
      descriptorIndex2=findDescriptorIndex(descriptor,
					   NumberConfigurations);
      
                if (descriptorIndex2 <= 0) continue;
//      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//	sum+=(-1.0)*CCMcoeffs[descriptorIndex1]
//	  *CCMcoeffs[descriptorIndex2]
//	  *hh[term].coefficient[index][1];
      
      if (Control.file) {
	if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
	  int indices[7]={0};
	  double dummy;
	  indices[0]=config_index;
	  indices[1]=57;
	  indices[2]=descriptorIndex1;
	  indices[3]=descriptorIndex2;
	  indices[4]=0;
	  indices[5]=0;
	  dummy=(-1.0)*hh[term].coefficient[index][1];
	  collect_term(indices,dummy,h_id);
	}
      }
      
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

// s^z s^z : all s 

double term58(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  
  /* 
     TERM 58: 
     
     2 F(k) . S(+)(k) . S(z)(k)

  */
  
  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int index;
  double sum;
  sum=0.0000;
  
  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];
  
  // get the multiplying coefficient correct first!

  multiplier=2.0;

  for(k=1;k<=length;k++) {
    
    findDescriptorOfCluster(length,configuration,descriptor);
    descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
    
    if (descriptorIndex != config_index) {
      printf("Term 58 Error: descriptorIndex (%d"
	") != config_index (%d)...\nExiting!!\n\n",
        descriptorIndex, config_index);
      exit(1);
    }
    
    index=configuration[k][4];      
    
    // note that no *CCMcoeffs[config_index] to find in sum
    // because it's the lhs.
    
//    sum+=multiplier*(-0.5*UnitCell.s2[index])*
//      hh[term].coefficient[index][1];
    
    // -0.5*UnitCell.s2[new_index] -- coefficient from \tilde s^z \tilde s^z
    
    if (Control.file) {
      int indices[7]={0};
      double dummy;
      indices[0]=config_index;
      indices[1]=58;
      indices[2]=config_index;
      indices[3]=0;
      indices[4]=0;
      indices[5]=0;
      dummy=multiplier*(-0.5*UnitCell.s2[index])*
	hh[term].coefficient[index][1];
            collect_term(indices,dummy,h_id);
    }
    
  }
  free(configuration);
  return sum;
}

double term59(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 59: 
     
     F(k) . S(+)(k) . S(+)(k).
     
  */


  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i, j,  l;
  int k1;
  int k2;
  int are_they_the_same_lattice_site;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int tempConfig[MAX_LSUBM+2];
  int counter;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) {
      configuration[i][j]=string[i][j];
      conf[i][j]=0;
    }
  
  multiplier=1.0;
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {
	
	for(l=1;l<5;l++) { conf[1][l]=configuration[k1][l]; } // F(m)
	for(l=1;l<5;l++) { tempConfig[l]=configuration[k2][l]; } // s(k)^+
	
	are_they_the_same_lattice_site=0;
	index=conf[1][4];
	
	if(tempConfig[1] == conf[1][1] &&
	   tempConfig[2] == conf[1][2] &&
	   tempConfig[3] == conf[1][3] &&
	   tempConfig[4] == conf[1][4])
	  are_they_the_same_lattice_site=1;      
	  
	// now load the rest into conf
	
	counter=1;
	for(i=1;i<=length;i++) {
	  if ((i != k2) && (i != k1)) {
	    counter++;
	    for(l=1;l<5;l++)
	      conf[counter][l]=configuration[i][l];
	  }
	}
	
	if (counter != (length-1)) {
	  printf("TERM 59: Error counter != (length-1)...\nExiting\n\n");
	  exit(1);
	}
	
	if (are_they_the_same_lattice_site) {
	  
	  findDescriptorOfCluster(length-1,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
//	  if (descriptorIndex > 0)
//	    sum+=multiplier*CCMcoeffs[descriptorIndex]*
//	      hh[term].coefficient[index][1];	
	  
	  if (Control.file) {
	    if (descriptorIndex > 0){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=59;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*hh[term].coefficient[index][1];
            collect_term(indices,dummy,h_id);
	    }
	  }
	  
	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

double term60(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 60: 
     
     H_{k,k} . S(+)(k) . S(+)(k).
     
  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);


  int i, j, l;
  int k1;
  int k2;
  int are_they_the_same_lattice_site;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int counter;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) {
      configuration[i][j]=string[i][j];
      conf[i][j]=0;
    }
  
  multiplier=1.0;
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {
	
	for(l=1;l<5;l++) { conf[1][l]=configuration[k1][l]; } // F(m)
	for(l=1;l<5;l++) { conf[2][l]=configuration[k2][l]; } // s(k)^+
	
	are_they_the_same_lattice_site=0;
	index=conf[1][4];
	
	if(conf[2][1] == conf[1][1] &&
	   conf[2][2] == conf[1][2] &&
	   conf[2][3] == conf[1][3] &&
	   conf[2][4] == conf[1][4])
	  are_they_the_same_lattice_site=1;      
	  
	// now load the rest into conf
	
	counter=2;
	for(i=1;i<=length;i++) {
	  if ((i != k2) && (i != k1)) {
	    counter++;
	    for(l=1;l<5;l++)
	      conf[counter][l]=configuration[i][l];
	  }
	}
	
	if (counter != length) {
	  printf("TERM 60: Error counter != length...\nExiting\n\n");
	  exit(1);
	}
	
	if (are_they_the_same_lattice_site) {
	  
	  findDescriptorOfCluster(length,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
//	  if (descriptorIndex > 0)
//	    sum+=multiplier*CCMcoeffs[descriptorIndex]*
//	      hh[term].coefficient[index][1];	
	  
	  if (Control.file) {
	    if (descriptorIndex > 0){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=60;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*hh[term].coefficient[index][1];
            collect_term(indices,dummy,h_id);
	    }
	  }
	  
	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

double term61(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 61: 

     F(k) . F(k) . S(+)(k) . S(+)(k).
     
  */

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int are_they_the_same_lattice_site;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; 
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k2][m];
	
	// loop over nearest neighbours
	
	are_they_the_same_lattice_site=0;
	index=conf1[1][4];

	if(conf2[1][1] == conf1[1][1] &&
	   conf2[1][2] == conf1[1][2] &&
	   conf2[1][3] == conf1[1][3] &&
	   conf2[1][4] == conf1[1][4])
	  are_they_the_same_lattice_site=1;      
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_the_same_lattice_site) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=1;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) {
	      if (spin[j]) {
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++)
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    findDescriptorOfCluster(conf1_len,
					  conf1,
					  descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex1 <= 0) continue;
	    findDescriptorOfCluster(conf2_len,
					  conf2,
					  descriptor);
	    
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex2 <= 0) continue;
	    // now find the multiplying cofficient.
	    	    
//	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//	      sum+=CCMcoeffs[descriptorIndex1]
//		  *CCMcoeffs[descriptorIndex2]*
//		hh[term].coefficient[index][1];
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=61;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=hh[term].coefficient[index][1];
		collect_term(indices,dummy,h_id);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term62(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 62: 
     
     S(+)(k)

     */
  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  
  int index;
  double sum;

  sum=0.0000;

  if (length == 1) {

    index=string[1][4];

    sum+=hh[term].coefficient[index][1];
    
    if (Control.file) {
      int indices[7]={0};
      double dummy;
      indices[0]=config_index;
      indices[1]=62;
      indices[2]=0;
      indices[3]=0;
      indices[4]=0;
      indices[5]=0;
      dummy=hh[term].coefficient[index][1];
      collect_term(indices,dummy,h_id); 
    }
      
  }
  return sum;
}

/* Joerg Sep05:
 *  new: collect_gse_term() now collects gse data too
 * this may conflict with NoDisk/MemoryOnly-Version
 * ToDo: estimate the size!
 */
// Joerg Jun08: obsolete
//void collect_gse_term(int indices[6],double dummy,int id){
//    collect_term(indices,dummy,id);
//}

/* now find terms which contribute to the ground-state energy */

double gse_term1(struct hamiltonian_defn hh[], int term)
{

 /* 
     GROUND-STATE ENERGY TERM 1: 
     
     S(z)(k) S(z)(m) 

     Wed Jul 10 BST 2002
     
  */

  int i,j,neighbour,index,new_index,(*configuration)[5];
  double sum, multiplier;

  sum=0.0;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    for(neighbour=1;neighbour<=hh[term].k_to_m_number[index];
	neighbour++) {
      
      new_index=hh[term].k_to_m[index][neighbour][3];
      
      multiplier=1.0;

      sum+=multiplier*(-0.5*UnitCell.s2[new_index])*(-0.5*UnitCell.s2[index])*
	hh[term].coefficient[index][neighbour]/UnitCell.number;
      
      // -0.5*UnitCell.s2[new_index] -- coefficient from \tilde s^z \tilde s^z
      
      {
	int indices[7]={0};
	double dummy;
	indices[0]=0;
        indices[1]=term;
	indices[2]=0;
        indices[3]=0;
	indices[4]=0;
        indices[5]=0;
	dummy=multiplier*
	  (-0.5*UnitCell.s2[new_index])*
	  (-0.5*UnitCell.s2[index])*
	  hh[term].coefficient[index][neighbour]
	  /UnitCell.number;
        collect_term(indices,dummy,h_id);
      } /* ToDo: collect? */
      
    }
  }
  free(configuration);
  return sum;
}
  
  
double gse_term2(struct hamiltonian_defn hh[], int term)
{

 /* 
     GROUND-STATE ENERGY TERM 2: 
     
     4 G_{k,m} S(z)(k) S(z)(m) 

     Wed Jul 10 BST 2002
     
  */

  int i, j, neighbour;
  int index, new_index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];

  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    for(neighbour=1;neighbour<=hh[term].k_to_m_number[index];
	neighbour++) {
      
      configuration[1][1]=0;      // x
      configuration[1][2]=0;      // y
      configuration[1][3]=0;      // z
      configuration[1][4]=index;  // uc
      
      new_index=hh[term].k_to_m[index][neighbour][3];
      
      configuration[2][1]=configuration[1][1]+
	hh[term].k_to_m[index][neighbour][0];
      configuration[2][2]=configuration[1][2]+
	hh[term].k_to_m[index][neighbour][1];
      configuration[2][3]=configuration[1][3]+
	hh[term].k_to_m[index][neighbour][2];
      configuration[2][4]=
	hh[term].k_to_m[index][neighbour][3];
      
      findDescriptorOfCluster(2,configuration,descriptor);
      descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
      
      multiplier=4.0;

//      if (descriptorIndex > 0)
//	sum+=multiplier*
//	  (-0.5*UnitCell.s2[new_index])*
//	  (-0.5*UnitCell.s2[index])
//	  *hh[term].coefficient[index][neighbour]
//	  *CCMcoeffs[descriptorIndex]
//	  /UnitCell.number;
      
      {
	if (descriptorIndex > 0) {
	  int indices[7]={0};
	  double dummy;
	  indices[0]=0;
          indices[1]=term;
	  indices[2]=descriptorIndex; // 2particle-nn-cluster
          indices[3]=0;
	  indices[4]=0;
          indices[5]=0;
	  dummy=multiplier*
	    (-0.5*UnitCell.s2[new_index])*
	    (-0.5*UnitCell.s2[index])*
	    hh[term].coefficient[index][neighbour]
	    /UnitCell.number;
          collect_term(indices,dummy,h_id);
	}
      }

    }
  }
  free(configuration);
  return sum;
}
 
double gse_term3(struct hamiltonian_defn hh[], int term)
{

 /* 
     GROUND-STATE ENERGY TERM 3: 
     
     4 F_{k} F_{m} S(z)(k) S(z)(m) 

     Wed Jul 10 BST 2002
     
  */

  int i, j, neighbour;
  int index, new_index;
  int descriptor1[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptor2[MAX_SPIN_LIST];
  int descriptorIndex2;
  double multiplier;
  int (*configuration1)[5];
  int (*configuration2)[5];

  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  configuration2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) {
      configuration1[i][j]=0;
      configuration2[i][j]=0;
    }

  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    for(neighbour=1;neighbour<=hh[term].k_to_m_number[index];
	neighbour++) {
      
      configuration1[1][1]=0;
      configuration1[1][2]=0;
      configuration1[1][3]=0;
      configuration1[1][4]=index;
      
      new_index=hh[term].k_to_m[index][neighbour][3];
      
      configuration2[1][1]=configuration1[1][1]+
	hh[term].k_to_m[index][neighbour][0];
      configuration2[1][2]=configuration1[1][2]+
	hh[term].k_to_m[index][neighbour][1];
      configuration2[1][3]=configuration1[1][3]+
	hh[term].k_to_m[index][neighbour][2];
      configuration2[1][4]=
	hh[term].k_to_m[index][neighbour][3];
      
      findDescriptorOfCluster(1,configuration1,descriptor1);
      descriptorIndex1=findDescriptorIndex(descriptor1,NumberConfigurations);

                if (descriptorIndex1 <= 0) continue;
      findDescriptorOfCluster(1,configuration2,descriptor2);
      descriptorIndex2=findDescriptorIndex(descriptor2,NumberConfigurations); 
     
                if (descriptorIndex2 <= 0) continue;
      multiplier=4.0;
      multiplier=multiplier*(-0.5*UnitCell.s2[new_index])*(-0.5*UnitCell.s2[index]);
      
//      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//	sum+=multiplier
//	  *hh[term].coefficient[index][neighbour]
//	  *CCMcoeffs[descriptorIndex1]
//	  *CCMcoeffs[descriptorIndex2]/UnitCell.number;
      
      {
	if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
	  int indices[7]={0};
	  double dummy;
	  indices[0]=0;
	  indices[1]=term;
	  indices[2]=descriptorIndex1; // single site 0 - config
	  indices[3]=descriptorIndex2; // single neighbour site - config
	  indices[4]=0;
	  indices[5]=0;
	  dummy=multiplier*
	    hh[term].coefficient[index][neighbour]
	    /UnitCell.number;
	  collect_term(indices,dummy,h_id);
	}
      }
      
    }
  }
  free(configuration1);
  free(configuration2);
  return sum;
}
  
double gse_term4(struct hamiltonian_defn hh[], int term)
{

 /* 
     GROUND-STATE ENERGY TERM 4: 
     
     -2 F_{m} S(z)(k) S(z)(m) 

     Wed Jul 10 BST 2002
     
  */

  int i, j, neighbour;
  int index, new_index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];

  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    for(neighbour=1;neighbour<=hh[term].k_to_m_number[index];
	neighbour++) {
      
      new_index=hh[term].k_to_m[index][neighbour][3];
      
      configuration[1][1]=configuration[1][1]+
	hh[term].k_to_m[index][neighbour][0];
      configuration[1][2]=configuration[1][2]+
	hh[term].k_to_m[index][neighbour][1];
      configuration[1][3]=configuration[1][3]+
	hh[term].k_to_m[index][neighbour][2];
      configuration[1][4]=
	hh[term].k_to_m[index][neighbour][3];
      
      findDescriptorOfCluster(1,configuration,descriptor);
      descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
      
      multiplier=-2.0;

//      if (descriptorIndex > 0)
//	sum+=multiplier*
//	  (-0.5*UnitCell.s2[new_index])*
//	  (-0.5*UnitCell.s2[index])
//	  *hh[term].coefficient[index][neighbour]
//	  *CCMcoeffs[descriptorIndex]/UnitCell.number;
      
      {
	if (descriptorIndex > 0) {
	  int indices[7]={0};
	  double dummy;
	  indices[0]=0;
          indices[1]=term;
	  indices[2]=descriptorIndex;
          indices[3]=0;
	  indices[4]=0;
          indices[5]=0;
	  dummy=multiplier*
	    (-0.5*UnitCell.s2[new_index])*
	    (-0.5*UnitCell.s2[index])*
	    hh[term].coefficient[index][neighbour]
	    /UnitCell.number;
	  collect_term(indices,dummy,h_id);
	}
      }

    }
  }
  free(configuration);
  return sum;
}
  

double gse_term5(struct hamiltonian_defn hh[], int term)
{

 /* 
     GROUND-STATE ENERGY TERM 5: 
     
     -2 F_{k} S(z)(k) S(z)(m) 

     Wed Jul 10 BST 2002
     
  */

  int i, j, neighbour;
  int index, new_index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];

  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    for(neighbour=1;neighbour<=hh[term].k_to_m_number[index];
	neighbour++) {
      
      configuration[1][1]=0;
      configuration[1][2]=0;
      configuration[1][3]=0;
      configuration[1][4]=index;

      new_index=hh[term].k_to_m[index][neighbour][3];
      
      findDescriptorOfCluster(1,configuration,descriptor);
      descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
      
      multiplier=-2.0;

//      if (descriptorIndex > 0)
//	sum+=multiplier*
//	  (-0.5*UnitCell.s2[new_index])*
//	  (-0.5*UnitCell.s2[index])
//	  *hh[term].coefficient[index][neighbour]
//	  *CCMcoeffs[descriptorIndex]
//	  /UnitCell.number;
      
      {
	if (descriptorIndex > 0) {
	  int indices[7]={0};
	  double dummy;
	  indices[0]=0;
          indices[1]=term;
	  indices[2]=descriptorIndex;
          indices[3]=0;
	  indices[4]=0;
          indices[5]=0;
	  dummy=multiplier*
	    (-0.5*UnitCell.s2[new_index])*
	    (-0.5*UnitCell.s2[index])*
	    hh[term].coefficient[index][neighbour]
	    /UnitCell.number;
	  collect_term(indices,dummy,h_id);
	}
      }

    }
  }
  free(configuration);
  return sum;
}
  
double gse_term6(struct hamiltonian_defn hh[], int term)
{

 /* 
     GROUND-STATE ENERGY TERM 6: Single Site
     
     S(z)(k) 

     Wed Jul 10 BST 2002
     
  */

  int i,j,index,(*configuration)[5];
  double sum, multiplier;

  sum=0.0;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    multiplier=1.0;
    
    sum+=multiplier*(-0.5*UnitCell.s2[index])*
      hh[term].coefficient[index][1]/UnitCell.number;
    
    {
      int indices[7]={0};
      double dummy;
      indices[0]=0;
      indices[1]=term;
      indices[2]=0;
      indices[3]=0;
      indices[4]=0;
      indices[5]=0;
      dummy=multiplier*
	(-0.5*UnitCell.s2[index])*
	hh[term].coefficient[index][1]
	/UnitCell.number;
      collect_term(indices,dummy,h_id);
    }
      
  }
  free(configuration);
  return sum;
}
  
double gse_term7(struct hamiltonian_defn hh[], int term)
{

 /* 
     GROUND-STATE ENERGY TERM 7: 
     
     -2 F_{m} S(z)(k) 

     Wed Jul 10 BST 2002
     
  */

  int i, j;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];

  double sum;
  sum=0.0000;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    configuration[1][1]=0;
    configuration[1][2]=0;
    configuration[1][3]=0;
    configuration[1][4]=index;
      
    findDescriptorOfCluster(1,configuration,descriptor);
    descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
    
    multiplier=-2.0;
    
//    if (descriptorIndex > 0)
//      sum+=multiplier*
//	(-0.5*UnitCell.s2[index])*
//	hh[term].coefficient[index][1]
//	*CCMcoeffs[descriptorIndex]/UnitCell.number;
    
    {
      if (descriptorIndex > 0) {
	int indices[7]={0};
	double dummy;
	indices[0]=0;
        indices[1]=term;
	indices[2]=descriptorIndex;
        indices[3]=0;
	indices[4]=0;
        indices[5]=0;
	dummy=multiplier*
	  (-0.5*UnitCell.s2[index])*
	  hh[term].coefficient[index][1]
	  /UnitCell.number;
	collect_term(indices,dummy,h_id);
      }
    }
    
  }
  free(configuration);
  return sum;
}

double gse_term8(struct hamiltonian_defn hh[], int term)
{

 /* 
     GROUND-STATE ENERGY TERM 8: Single Site
     
     (S(z)(k))^2 

     Wed Jul 10 BST 2002
     
  */

  int i,j,index,(*configuration)[5];
  double sum, multiplier;

  sum=0.0;

  if (!hh[term].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    multiplier=1.0;
    
    sum+=multiplier*(-0.5*UnitCell.s2[index])*(-0.5*UnitCell.s2[index])*
      hh[term].coefficient[index][1]/UnitCell.number;
    
    {
      int indices[7]={0};
      double dummy;
      indices[0]=0;
      indices[1]=term;
      indices[2]=0;
      indices[3]=0;
      indices[4]=0;
      indices[5]=0;
      dummy=multiplier*
	(-0.5*UnitCell.s2[index])*
	(-0.5*UnitCell.s2[index])*
	hh[term].coefficient[index][1]
	/UnitCell.number;
      collect_term(indices,dummy,h_id);
    }
      
  }
  free(configuration);
  return sum;
}

int flag_subset=0; // Apr08 JS for SpeedUp

// same as obtainGenExp() ToDo: replace by one unique function!
void obtain(void)
{

  int    i,j,bonds;
  int    length;
  int    (*configuration)[5];
  int    dummy;
  double energy;
  double lhs, rhs;
//  char output[60];
  time_t t1, t2;
  struct hamiltonian_defn *hh;  /* [MAX_TERMS_h=63] */
  
//  static int n_out=0;   /* number of outputs */
#define MAX_out 10
  
//  char gs_name[100]="";

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  flag_subset=1;  // do subset-Check before findDescriptorOfCluster Apr08 JS

  if (!myid) printf("\n");

  // The data structures binary_max[], trimary_max[] ... are used 
  // to partition a given cluster into two (e.g. F(k) F(m)), three 
  // .... pieces in each `term' in H. 
  
  binary_max[0]=1;     dummy=2;   // 2^0 .. 2^30  # 30bit
  for(i=1;i<=30;i++) { binary_max[i]=dummy;   dummy=2*dummy; }

  trimary_max[0]=1;    dummy=3;   // 3^0 .. 3^10  # 16bit, i<=MAX_LSUBM
  for(i=1;i<=10;i++) { trimary_max[i]=dummy;  dummy=3*dummy; }

  quartary_max[0]=1;   dummy=4;   // 4^0 .. 4^10  # 20bit
  for(i=1;i<=10;i++) { quartary_max[i]=dummy; dummy=4*dummy; }

//  strcpy(output, Control.name);  strcat(output,".terms"); /* terms */
//  strcpy(gs_name,Control.name);  strcat(gs_name,"_gs");   /* gs terms */

  // terms_fp=fopen(output,"w"); => done in ccm.C now

  // ground-state energy equation.

  // fp_gs=fopen(gs_name,"w");  /* same content for all threads */


  /* NEW!!: work out the maximum radius for all clusters 
     This is used to reduce the amount of sweeping for 
     F_k F_m s_k^z s_m^z -like terms. 
      v060611 JS: this function is moved here from input and define (copy)
  */
  radius=0;
  {
    int length;
    int (*tempConfig)[5]=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
    for(i=1;i<=NumberConfigurations;i++) {
      length=findClusterFromDescriptor(spinConfigurations[i].descriptor,
                                       tempConfig);
      /* note that midPointX, midPointY, midPointZ indicate Unit cells */
      // note that integer rounding shouldn't affect this 
      // algorithm
      // now find the maximum `radius'
      for(j=1;j<=length;j++) {
        if (abs(tempConfig[j][1]) > radius) radius=abs(tempConfig[j][1]);
        if (abs(tempConfig[j][2]) > radius) radius=abs(tempConfig[j][2]);
        if (abs(tempConfig[j][3]) > radius) radius=abs(tempConfig[j][3]);
      }
    }
    radius+=1;  // ToDo: simpler maxdist m? or compute for each config (smaller)?
    if (!myid)
      printf("\nmaximum cell radius of all clusters = %d\n\n",radius);
    free(tempConfig);
  }

  // find CCM equations.

  if (!myid){
    printf(" Write: %s_gs %s.terms (buffered)\n\n",Control.name,Control.name);
    printf("==========================================================\n\n");
    printf("OBTAIN CCM EQUATIONS: START ALGORITHM.\n\n");
    printf("==========================================================\n\n");
  }

  lhs=0.0;rhs=0.0;energy=0.0;
  /* delta=1.0;  Joerg Sep05: delta seems to be not used here ??? */

  h_id=0; // used for collect_term

  for(bonds=1;bonds<=Params.bonds;bonds++) {

    hh=Hamiltonian[bonds];  // abbreviation
    
    // find ground-state energy
    energy += /* delta * */ gse_term1(hh, 1); // term13  1 
    energy += /* delta * */ gse_term2(hh,27); // term13  9
    energy += /* delta * */ gse_term3(hh,30); // term13  9
    energy += /* delta * */ gse_term4(hh, 8); // term13  3
    energy += /* delta * */ gse_term5(hh,17); // term13  5
    energy += /* delta * */ gse_term6(hh,54); // term13  7
    energy += /* delta * */ gse_term7(hh,55); // term13 10
    energy += /* delta * */ gse_term8(hh,58); // term13 11
    {
      int indices[7]={0};
      indices[0]=STOP_CFG;indices[1]=STOP_TERM;  /* mark end of bond data */
      indices[2]=STOP_CFG;indices[3]=STOP_CFG;
      indices[4]=STOP_CFG;indices[5]=STOP_CFG;
      collect_term(indices,0.0,0);
    } 
  }

  // fclose(fp_gs);

  /* added for performance analysis, Joerg 2005 */
  if (Control.output > 1 && myid == 0)
  for(int term=1; term<=62; term++)
  for(int cell=1; cell<=UnitCell.number; cell++) {
    int bonds;
    for (bonds=1; bonds<=Params.bonds; bonds++)
      if (Hamiltonian[bonds][term].evaluate) break;
    if (bonds>Params.bonds) continue; // skip zero-terms
    printf("DBG: term62 term13 cell neighbours[bond] = %2d %2d %2d",
           term, term_13[term], cell);
    for (bonds=1; bonds<=Params.bonds; bonds++)
       printf(" %2d", Hamiltonian[bonds][term].k_to_m_number[cell]);
    printf("\n");
  }

#if 0   // never used
  /* Joerg: where is CCMcoeffs set? check it! */
  for (j=0, i=0; i<MAX_NUM_CONFIGURATIONS; i++)
    if (CCMcoeffs[i]) { 
       j++;
       printf("%d: CCMcoeffs[%d]= %f\n", myid, i, CCMcoeffs[i]);
    }
  if (j) printf("%d: # nonzero CCMcoeffs: %d\n", myid, j);
#endif

  /* This is very slow (only for bond2!) (Joerg 2005, ToDo) */
  t1=time(NULL);
  h_id=1; // used for collect_term

  for(bonds=1;bonds<=Params.bonds;bonds++) {

    // n_out=0;  // reset for each bond    

    for(i=1;i<=NumberConfigurations;i++) { // total number of clusters

      /* not well balanced / scales badly (for 1D only, small #maxlencfgs)
       *       cpu time is not the same for each config -> balancing?
       * ToDo: one global counter for all threads, JS06
       *       let only one thread pass
       *       thread 0 starts with biggest config (=last cfg?)
       *       send counter-1 to next thread unblocked and start calculation
       *       receive counter and send -1 to next counter, stop if counter<0
       *       - send token to next ready process
       *
       *  output terms_per_second about every minute
       *    or on regular outputs
       */
      if (((i%npe)==myid) || PARALLEL_FLAG==0) {
        
        static unsigned long nt_print=1; // used to output all 4^n

	length = findClusterFromDescriptor(spinConfigurations[i].descriptor,
		   			   configuration);        
	
        // we want to output not to much because not every user knows
        //    what he does ;)
        // ToDo: output only for numterms=1,4,16,64,256,1024,4096,...
	if (Control.output > 0 && myid == 0
         && nterms_out1 > nt_print) {  // we only output every 4^n
          t2=time(NULL)-t1;
          if (nt_print<<1) nt_print<<=1;  // *4 but avoid overflow
          if (t2>27)  // Apr08 reduce output 
          printf("obtain: Bond %d : Cluster<= %5d"
                 " : terms+= %8ld %5ld/s %5.1fm\n",
                 bonds, i, nterms_out1, ((t2>0)?nterms_out1/t2:0), t2/60. );
          fflush(stdout); // no buffering (possible time estimations)
          // Joerg: its already printed out on define.C 
	  // printOutDescriptor(spinConfigurations[i].descriptor);
	  // printOutClusterFromDescriptor(spinConfigurations[i].descriptor);
          if (i==MAX_out)
            printf("obtain: output is reduced after %d terms\n",MAX_out);
        }

        hh=Hamiltonian[bonds];  // abbreviation
	
	// s^z s^z : s all s = t1
	
	lhs += term2(i,length,configuration,hh,2);  // term_13[2]=1
	lhs += term3(i,length,configuration,hh,3);
	lhs += term4(i,length,configuration,hh,4);
	lhs += term5(i,length,configuration,hh,5);
	
	// s^z s^+ all s = t2
	
	rhs += term6(i,length,configuration,hh,6);  // term_13[6]=2
	rhs += term7(i,length,configuration,hh,7);
      
	// s^z s^- all s = t3
	
	rhs += term8(i,length,configuration,hh,8); // term_13[8]=3
	rhs += term9(i,length,configuration,hh,9);
	rhs += term10(i,length,configuration,hh,10);
	rhs += term11(i,length,configuration,hh,11);
	rhs += term12(i,length,configuration,hh,12);
	rhs += term13(i,length,configuration,hh,13);
	
	// s^+ s^z all s = t4
	
	rhs += term14(i,length,configuration,hh,14); // term_13[14]=4
	rhs += term15(i,length,configuration,hh,15);
      
      // s^- s^z all s
      
	rhs += term16(i,length,configuration,hh,16);
	rhs += term17(i,length,configuration,hh,17);
	rhs += term18(i,length,configuration,hh,18);
	rhs += term19(i,length,configuration,hh,19);
	rhs += term20(i,length,configuration,hh,20);
	rhs += term21(i,length,configuration,hh,21);

      // s^+ s^- all s

	rhs += term22(i,length,configuration,hh,22);
	rhs += term23(i,length,configuration,hh,23);

      // s^- s^+ all s

	rhs += term24(i,length,configuration,hh,24);
	rhs += term25(i,length,configuration,hh,25);

      
      // s^+ s^+ all s = t8

	rhs += term26(i,length,configuration,hh,26); // term_13[26]=8
      
      // s^- s^- all s

	rhs += term27(i,length,configuration,hh,27); /* => length+1 */
	rhs += term28(i,length,configuration,hh,28);
	rhs += term29(i,length,configuration,hh,29);
	rhs += term30(i,length,configuration,hh,30); /* slowest 75% square */
	rhs += term31(i,length,configuration,hh,31); /* t30..34 90% of slowness */
	rhs += term32(i,length,configuration,hh,32);
	rhs += term33(i,length,configuration,hh,33);
	rhs += term34(i,length,configuration,hh,34); /* slowest 31% chain */
	rhs += term35(i,length,configuration,hh,35);

      // s^- s^- s > 1/2  = t9
      
	rhs += term36(i,length,configuration,hh,36); // term_13[36]=9
	rhs += term37(i,length,configuration,hh,37);
	rhs += term38(i,length,configuration,hh,38);
	rhs += term39(i,length,configuration,hh,39);
	rhs += term40(i,length,configuration,hh,40);
	rhs += term41(i,length,configuration,hh,41);
	rhs += term42(i,length,configuration,hh,42);
	rhs += term43(i,length,configuration,hh,43);
	rhs += term44(i,length,configuration,hh,44);
	rhs += term45(i,length,configuration,hh,45);
      
      // s^z s^- s > 1/2  = t3

	rhs += term46(i,length,configuration,hh,46); // term_13[46]=3
	rhs += term47(i,length,configuration,hh,47);
	rhs += term48(i,length,configuration,hh,48);

      // s^- s^z s > 1/2  = t5

	rhs += term49(i,length,configuration,hh,49); // term_13[49]=5
	rhs += term50(i,length,configuration,hh,50);
	rhs += term51(i,length,configuration,hh,51);

      // s^+ s^- s > 1/2 = t6
     
	rhs += term52(i,length,configuration,hh,52); // term_13[52]=6

      // s^- s^+ s > 1/2 = t7
     
	rhs += term53(i,length,configuration,hh,53); // term_13[53]=7

      // s^z all s = t10

	rhs += term54(i,length,configuration,hh,54); // term_13=10
     
      // s^- all s = t11

	rhs += term55(i,length,configuration,hh,55); // term_13=11
	rhs += term56(i,length,configuration,hh,56);
	rhs += term57(i,length,configuration,hh,57);

      // (s^z)^2 = t12

	rhs += term58(i,length,configuration,hh,58); // term_13=12
	rhs += term59(i,length,configuration,hh,59);
	rhs += term60(i,length,configuration,hh,60);
	rhs += term61(i,length,configuration,hh,61);

      // s^+     = t13

	rhs += term62(i,length,configuration,hh,62); // term_13=13
      }
    }
    if (Control.file) {
      int indices[7]={0};
      indices[0]=STOP_CFG;indices[1]=STOP_TERM;
      indices[2]=STOP_CFG;indices[3]=STOP_CFG;
      indices[4]=STOP_CFG;indices[5]=STOP_CFG;
      collect_term(indices,0.0,1);
    }
    if (Control.output>0 && myid==0)
      printf("END H[bonds=%d] ==================================\n\n",bonds);
  }
  
  t2=time(NULL)-t1; // ToDo: min/max(MPI)
  if (myid==0) {
    printf("\nOBTAIN CCM EQUATIONS: SUCCESSFULLY COMPLETED!\n\n");
    printf(" obtain.t/min=%.2f (jobid=0)", t2/60.);
  }

  // fclose(terms_fp); => closed in ccm.C now

  free(configuration); // memleak until v0606

}
